from functools import reduce, update_wrapper
from typing import Optional, Any
from os import environ
from sys import stdin
import sys  # Testability requires sys.exit() instead of just exit()

from traceback import format_exc
from xml.etree.ElementTree import fromstring, ParseError

from pytos2.utils import setup_logger
from pytos2.securechange import Scw
from .ticket import Ticket

LOGGER = setup_logger("trigger")

SCW_EVENT = "SCW_EVENT"
# XXX SCW_EVENT is test. so just use that instead of trying to read in stdin first. -_-


class Trigger:
    CURRENT_TRIGGER = None
    TRIGGERS: set = set()
    TICKET_INFO = None
    TICKET = None

    _bits = 0b1

    @staticmethod
    def from_scw_event(trigger):
        for t in Trigger.TRIGGERS:
            if t == trigger:
                return t

        return NONE

    @classmethod
    def _get_bits(cls):
        old_bits = cls._bits

        cls._bits *= 0b10

        return old_bits

    def __init__(self, representation, string, bits=None):
        if bits is None:
            bits = Trigger._get_bits()

        if type(representation) is str:
            representation = [representation]

        if type(string) is str:
            string = [string]

        self._bits = bits
        self._representation = representation
        self._string = string

    def __eq__(self, other):
        if type(other) is str:
            return repr(self) == other
        elif isinstance(other, Trigger):
            return self._bits == other._bits
        else:
            return False

    def __contains__(self, item):
        if not hasattr(item, "_bits"):
            return False

        return self._bits & item._bits == item._bits

    def __or__(self, other):
        if other in self:
            return self  # XXX do we want to return a copied version of that trigger?, or the actual trigger?

        if self in other:
            return other

        # FIXME: Is this check needed? other in self and self in other should take care of it?
        if self._bits == 0 or other._bits == 0:
            return NONE  # noqa

        # FIXME: I believe this check is covered by `other in self` and `self in other` implicitly
        if self._bits == other._bits:
            return self  # noqa

        # XXX make the reprs and strings a set, not a list, so we don't have to
        # do the above if checks
        return Trigger(
            self._representation + other._representation,
            self._string + other._string,
            self._bits | other._bits,
        )

    def __str__(self):
        if len(self._string) == 1:
            return self._string[0] + " Trigger"
        else:
            return " or ".join(self._string) + " Triggers"

    def __repr__(self):
        return "|".join(self._representation)

    def __hash__(self):
        return hash(repr(self))


TEST = Trigger("TEST", "Test")
CREATE = Trigger("CREATE", "Create")
CLOSE = Trigger("CLOSE", "Close")
CANCEL = Trigger("CANCEL", "Cancel")
REJECT = Trigger("REJECT", "Reject")
ADVANCE = Trigger("ADVANCE", "Advance")
REDO = Trigger("REDO", "Re-do")
RESUBMIT = Trigger("RESUBMIT", "Re-submit")
REOPEN = Trigger("REOPEN", "Re-open")
AUTOMATION_FAILED = Trigger("AUTOMATION_FAILED", "Automation Failed")
PRE_ASSIGNMENT_SCRIPT = Trigger("PRE_ASSIGNMENT_SCRIPT", "Pre-assignment script")
NONE = Trigger("NONE", "No", 0)
ALL = (
    CREATE
    | CLOSE
    | CANCEL
    | REJECT
    | ADVANCE
    | REDO
    | RESUBMIT
    | REOPEN
    | AUTOMATION_FAILED
    | PRE_ASSIGNMENT_SCRIPT
)

# the list of triggers received by SecureChange directly.
Trigger.TRIGGERS = {
    TEST,
    CREATE,
    CLOSE,
    CANCEL,
    REJECT,
    ADVANCE,
    REDO,
    RESUBMIT,
    REOPEN,
    AUTOMATION_FAILED,
}

STDIN = None


def _get_ticket_info_from_stdin():
    global STDIN
    # XXX get rid of the closure, and just use STDIN
    std_info = None

    def _save_ticket_info_from_stdin():
        nonlocal std_info
        global STDIN

        if std_info is not None:
            return std_info
        elif stdin.isatty():
            return None
        else:
            STDIN = stdin.read()
            # if we consumed stdin wrongly, then save it in STDIN for later consumers (like cron jobs)
            try:
                std_info = fromstring(STDIN)

                # then we're not a ticket_info XML
                if std_info.tag != "ticket_info":
                    return None

            except ParseError:
                # Then we're not XML
                return None

            return std_info

    return _save_ticket_info_from_stdin


get_ticket_info_from_stdin = _get_ticket_info_from_stdin()


def get_ticket_id_from_ticket_info(ticket_info):
    try:
        text_id = ticket_info.find("id").text
    except AttributeError:
        return None

    return int(text_id)


# FIXME: Do we need this function for pytos2?
def get_workflow_name(ticket):  # noqa
    return ticket["workflow"]["name"]


def workflow_is(workflow, ticket=None):
    def _workflow_is(ticket):
        print(ticket.workflow)

        if ticket is None:
            return False
        if ticket.workflow is None:
            return False

        return workflow == ticket.workflow.name

    return _workflow_is if ticket is None else _workflow_is(ticket)


# XXX alternatively, accept an integer step.
# step_is works on the current step. so "on advance, when the step we advanced to is X, then do this".
# use previous step to talk about the previous step.
def step_is(step, ticket=None):
    def _step_is(ticket):
        if ticket is None:
            return False

        # We use `.get` because completed tickets don't have current steps.
        # use get_current_step XXX
        # rename the "get step" and "get task" functions to match
        return step == ticket.current_step_name

    return _step_is if ticket is None else _step_is(ticket)


def rejected_on(step, ticket=None):
    def _rejected_on(ticket):
        if ticket is None:
            return False

        if len(ticket.comments) == 0:
            return False

        return step == ticket.comments[-1].task_name

    return _rejected_on if ticket is None else _rejected_on(ticket)


# XXX the name of this function might be deceiving? it only matches failures on
# the recent-most step


def failure_was(failure, ticket=None):
    def _failure_was(ticket):
        if ticket is None:
            return False

        return failure == ticket.get_last_automation_failure_on_current_step()

    return _failure_was if ticket is None else _failure_was(ticket)


def get_previous_stage(ticket):
    return ticket.get_step(-2).name


def get_current_stage(ticket_info):
    current_stage = ticket_info.find("current_stage")

    if current_stage is None:
        return None

    current_stage_name = current_stage.find("name")

    if current_stage_name is None:
        return None
    else:
        return current_stage_name.text


def triggers_are_racing(trigger, ticket_info, ticket):
    if trigger not in (CREATE, ADVANCE):
        # XXX I was told that tickets can only race on the CREATE and ADVANCE
        # triggers, so fast fail if that's not the case
        return False

    ticket_info_current_stage = get_current_stage(ticket_info)
    ticket_previous_stage = get_previous_stage(ticket)

    if ticket_info_current_stage != ticket_previous_stage:
        LOGGER.warning(
            f"Race detected. Ticket stage in trigger ({ticket_info_current_stage}) does not match previous stage in the retrieved ticket ({ticket_previous_stage})"
        )

    return ticket_info_current_stage != ticket_previous_stage


def prime_from_securechange_trigger():
    # if Trigger.TICKET:  # XXX turn Trigger.TICKET into the special ticket object
    #     return (Trigger.CURRENT_TRIGGER, Trigger.TICKET_INFO, Trigger.TICKET)

    ticket_info = get_ticket_info_from_stdin()
    # TODO CLEANUP
    #     if Trigger.ENTRYPOINT is None:
    #         LOGGER.error("No Scw instance initialized")
    #         return (None, None, None)
    if ticket_info is None:
        LOGGER.warning("Got empty ticket_info")
        return (None, None, None)

    trigger = Trigger.from_scw_event(environ["SCW_EVENT"])
    LOGGER.debug(f"Got {trigger} event")

    if trigger == TEST:
        LOGGER.debug("Got test trigger")
        return (trigger, ticket_info, None)

    ticket_id = get_ticket_id_from_ticket_info(ticket_info)
    LOGGER.debug("trigger found ticket_id: {}".format(ticket_id))

    if ticket_id is None:
        # XXX not sure why this would happen.
        return (None, None, None)

    ticket = Scw.default.get_ticket(ticket_id)

    if triggers_are_racing(trigger, ticket_info, ticket):
        LOGGER.debug("Detected race condition")
        sys.exit(0)

    return (trigger, ticket_info, ticket)


def _listify(xs):
    return xs if isinstance(xs, list) else [xs]


def when(*_predicates, **kwargs):
    if Trigger.CURRENT_TRIGGER is None:
        (
            Trigger.CURRENT_TRIGGER,
            Trigger.TICKET_INFO,
            Trigger.TICKET,
        ) = prime_from_securechange_trigger()
    # this seems unecessary
    # predicates = reduce(lambda ps, p: ps + _listify(p), _predicates, [])

    def _when(f):
        def _when2(*f_args, **f_kwargs):
            return f(*f_args, **f_kwargs)

        if Trigger.CURRENT_TRIGGER is None:
            pass
        elif not Trigger.TICKET:
            pass
        else:
            try:
                should_run = any(
                    ps(Ticket.kwargify(Trigger.TICKET.data)) for ps in _predicates
                )
            except Exception:
                LOGGER.error(format_exc())
                sys.exit(1)

            if should_run:
                try:
                    _when2(Trigger.TICKET, **kwargs)
                except Exception:
                    LOGGER.error(format_exc())
                    sys.exit(1)

        return update_wrapper(_when2, f)

    return _when


def on(*_trigger, when=lambda t: True, **kwargs):
    # duplicate with when because we need to prime
    # XXX refactor later
    (
        Trigger.CURRENT_TRIGGER,
        Trigger.TICKET_INFO,
        Trigger.TICKET,
    ) = prime_from_securechange_trigger()

    # on without a parameter runs always
    if len(_trigger) == 1 and callable(_trigger[0]):
        try:
            _trigger[0](Trigger.TICKET, **kwargs)
        except Exception:
            LOGGER.error(format_exc())
            sys.exit(1)

        return _trigger[0]

    trigger = None

    def _on(f):
        if (trigger is None) or (Trigger.CURRENT_TRIGGER in trigger):
            return globals()["when"](when, **kwargs)(f)
        else:
            return f

    for t in _trigger:
        if type(t) is list:
            if trigger is None:
                trigger = reduce(lambda ts, t: ts | t, t[1:], t[0])
            else:
                trigger = reduce(lambda ts, t: ts | t, t, trigger)
        else:
            if trigger is None:
                trigger = t
            else:
                trigger |= t

    return _on


def on_test(*args, when=lambda t: True, **kwargs):
    if args:
        return on(TEST, when=when, **kwargs)(args[0])
    else:
        return on(TEST, when=when, **kwargs)


def on_create(*args, when=lambda t: True, **kwargs):
    if args:
        return on(CREATE, when=when, **kwargs)(args[0])
    else:
        return on(CREATE, when=when, **kwargs)


def on_close(*args, when=lambda t: True, **kwargs):
    if args:
        return on(CLOSE, when=when, **kwargs)(args[0])
    else:
        return on(CLOSE, when=when, **kwargs)


def on_cancel(*args, when=lambda t: True, **kwargs):
    if args:
        return on(CANCEL, when=when, **kwargs)(args[0])
    else:
        return on(CANCEL, when=when, **kwargs)


def on_reject(*args, when=lambda t: True, **kwargs):
    if args:
        return on(REJECT, when=when, **kwargs)(args[0])
    else:
        return on(REJECT, when=when, **kwargs)


def on_advance(*args, when=lambda t: True, **kwargs):
    if args:
        return on(ADVANCE, when=when, **kwargs)(args[0])
    else:
        return on(ADVANCE, when=when, **kwargs)


def on_redo(*args, when=lambda t: True, **kwargs):
    if args:
        return on(REDO, when=when, **kwargs)(args[0])
    else:
        return on(REDO, when=when, **kwargs)


def on_resubmit(*args, when=lambda t: True, **kwargs):
    if args:
        return on(RESUBMIT, when=when, **kwargs)(args[0])
    else:
        return on(RESUBMIT, when=when, **kwargs)


def on_reopen(*args, when=lambda t: True, **kwargs):
    if args:
        return on(REOPEN, when=when, **kwargs)(args[0])
    else:
        return on(REOPEN, when=when, **kwargs)


def on_automation_failure(*args, when=lambda t: True, **kwargs):
    if args:
        return on(AUTOMATION_FAILED, when=when, **kwargs)(args[0])
    else:
        return on(AUTOMATION_FAILED, when=when, **kwargs)


def on_all(*args, when=lambda t: True, **kwargs):
    if args:
        return on(ALL, when=when, **kwargs)(args[0])
    else:
        return on(ALL, when=when, **kwargs)



from enum import Enum
from typing import Optional, List, Union

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop


class ServiceXsiType(Enum):
    SERVICE_GROUP_OBJECT = "serviceGroupDTO"
    SINGLE_SERVICE_OBJECT = "singleServiceDTO"


@propify
class Service(Jsonable):
    class ClassName(Enum):
        # classname and # type map
        ANY_OBJECT = "any_object"  # ip_service
        TCP_SERVICE = "tcp_service"  # tcp_service
        ICMP_SERVICE = "icmp_service"  # icmp_service
        OTHER_SERVICE = "other_service"  # ip_service, other_service
        UDP_SERVICE = "udp_service"  # udp_service
        SERVICE_GROUP = "service_group"  # group
        RPC_SERVICE = "rpc_service"  # rpc_service
        DCERPC_SERVICE = "dcerpc_service"  # other_service
        COMPOUND_TCP_SERVICE = "compound_tcp_service"  # other_service
        GTP_MM_V1_SERVICE = "gtp_mm_v1_service"  # other_service
        GTP_MM_V0_SERVICE = "gtp_mm_v0_service"  # other_service
        GTP_SERVICE = "gtp_service"  # other_service
        GTP_V1_SERVICE = "gtp_v1_service"  # other_service
        ICMP_V6_SERVICE = "icmp_v6_service"  # other_service (no min/max port)
        ICMPV6_SERVICE = "icmpv6_service"  # ip_service (has min/max port)
        PORT_SERVICE = "port_service"  # port_service
        TCP_CITRIX_SERVICE = "tcp_citrix_service"  # other_service
        SERVICE = "service"  # other_service

    class Type(Enum):
        TCP_SERVICE = "tcp_service"
        IP_SERVICE = "ip_service"
        GROUP = "group"
        ICMP_SERVICE = "icmp_service"
        UDP_SERVICE = "udp_service"
        OTHER_SERVICE = "other_service"
        RPC_SERVICE = "rpc_service"
        PORT_SERVICE = "port_service"

    class Protocol(Enum):
        TCP = 6
        UDP = 17
        ICMP = 13

    class Prop(Enum):
        XSI_TYPE = "@xsi.type"
        GLOBAL = "global"
        MIN = "min"
        MAX = "max"
        IMPLICIT = "implicit"
        ICMP_CODE = "icmp_code"

    xsi_type: Optional[ServiceXsiType] = prop(
        None, key=Jsonable.Prop.XSI_TYPE.value, repr=False
    )
    name: Optional[str] = prop(None)
    display_name: Optional[str] = prop(None)
    class_name: Optional[ClassName] = prop(None, repr=False)
    type: Optional[Type] = prop(None, repr=False)
    is_global: bool = prop(False, key=Prop.GLOBAL.value, repr=False)
    comment: Optional[str] = prop(None, repr=False)
    uid: Optional[str] = prop(None, repr=False)
    overrides: bool = prop(False, repr=False)
    negate: bool = prop(False, repr=False)
    timeout: Optional[Union[str, int]] = prop(None, repr=False)
    is_implicit: bool = prop(False, repr=False, key=Prop.IMPLICIT.value)

    device_id: Optional[int] = prop(None, repr=False)


@propify
class ServiceObjectReference(Jsonable):
    uid: Optional[str] = prop(None)
    name: Optional[str] = prop(None)
    display_name: Optional[str] = prop(None)


@propify
class TCPServiceObject(Service):
    protocol: Optional[int] = prop(None)
    min_port: Optional[int] = prop(None, key=Service.Prop.MIN.value)
    max_port: Optional[int] = prop(None, key=Service.Prop.MAX.value)


@propify
class UDPServiceObject(Service):
    protocol: Optional[int] = prop(None, converter=optional(int))
    min_port: Optional[int] = prop(
        None, converter=optional(int), key=Service.Prop.MIN.value
    )
    max_port: Optional[int] = prop(
        None, converter=optional(int), key=Service.Prop.MAX.value
    )


@propify
class ICMPServiceObject(Service):
    protocol: Optional[int] = prop(
        None, converter=optional(int), key=Service.Prop.ICMP_CODE.value
    )
    min_port: Optional[int] = prop(
        None, converter=optional(int), key=Service.Prop.MIN.value
    )
    max_port: Optional[int] = prop(
        None, converter=optional(int), key=Service.Prop.MAX.value
    )


@propify
class IPServiceObject(Service):
    min_port: Optional[int] = prop(
        None, converter=optional(int), key=Service.Prop.MIN.value
    )
    max_port: Optional[int] = prop(
        None, converter=optional(int), key=Service.Prop.MAX.value
    )


@propify
class PortServiceObject(Service):
    min_port: Optional[int] = prop(
        None, converter=optional(int), key=Service.Prop.MIN.value
    )
    max_port: Optional[int] = prop(
        None, converter=optional(int), key=Service.Prop.MAX.value
    )


@propify
class ICMPV6ServiceObject(Service):
    management_domain_securetrack_name: Optional[str] = prop(None)
    management_domain: Optional[str] = prop(None)


@propify
class ICMPV6IPServiceObject(IPServiceObject, ICMPV6ServiceObject):
    pass


@propify
class OtherServiceObject(Service):
    pass


@propify
class OtherIPServiceObject(OtherServiceObject, IPServiceObject):
    pass


@propify
class AnyObject(Service):
    pass


@propify
class AnyIPServiceObject(AnyObject, IPServiceObject):
    pass


@propify
class ServiceGroup(Service):
    class Prop(Enum):
        MEMBER = "member"

    members: List[ServiceObjectReference] = prop(
        key=Prop.MEMBER.value, factory=list, repr=False
    )


@propify
class RPCServiceObject(Service):
    pass


@propify
class DCERPCService(Service):
    pass


@propify
class DefaultService(Service):
    pass


def classify_service_object(obj):
    # Handle a raer case where the singleServiceDTO result is missing a class_name
    key_name = "class_name" if "class_name" in obj else "type"

    if obj[key_name] == Service.ClassName.TCP_SERVICE.value:
        return TCPServiceObject.kwargify(obj)
    elif obj[key_name] == Service.ClassName.UDP_SERVICE.value:
        return UDPServiceObject.kwargify(obj)
    elif obj[key_name] == Service.ClassName.ICMP_SERVICE.value:
        return ICMPServiceObject.kwargify(obj)
    elif obj[key_name] == Service.ClassName.SERVICE_GROUP.value:
        return ServiceGroup.kwargify(obj)
    elif obj[key_name] == Service.ClassName.OTHER_SERVICE.value:
        if obj["type"] == Service.Type.IP_SERVICE.value:
            return OtherIPServiceObject.kwargify(obj)
        elif obj["type"] == Service.Type.OTHER_SERVICE.value:
            return OtherServiceObject.kwargify(obj)
    elif obj[key_name] == Service.ClassName.RPC_SERVICE.value:
        return RPCServiceObject.kwargify(obj)
    elif obj[key_name] == Service.ClassName.DCERPC_SERVICE.value:
        return DCERPCService.kwargify(obj)
    elif obj[key_name] == Service.ClassName.ANY_OBJECT.value:
        if obj["type"] == Service.Type.IP_SERVICE.value:
            return AnyIPServiceObject.kwargify(obj)
    elif obj[key_name] == Service.ClassName.PORT_SERVICE.value:
        return PortServiceObject.kwargify(obj)
    elif obj[key_name] == Service.ClassName.ICMPV6_SERVICE.value:
        return ICMPV6IPServiceObject.kwargify(obj)
    elif obj[key_name] == Service.ClassName.ICMP_V6_SERVICE.value:
        return ICMPV6ServiceObject.kwargify(obj)
    else:  # ultimate safe catch
        return DefaultService.kwargify(obj)  # pragma: no cover



import pytest
import json
import responses

from pytos2.securetrack.generic_route import GenericRoute
from pytos2.utils import get_api_node


class TestGenericRoute:
    mgmt = json.load(open("tests/securetrack/json/generic_routes/mgmt-3.json"))
    routes = mgmt["GenericRoutes"]

    route27 = json.load(open("tests/securetrack/json/generic_routes/route-27.json"))
    route = route27["GenericRoute"]

    @responses.activate
    def test_add_generic_routes_200(self, st):
        """Add one or multiple routes"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/route",
            status=201,
        )

        """Add single route"""
        newRoute = st.add_generic_route(
            device=self.route["mgmtId"],
            destination=self.route["destination"],
            mask=self.route["mask"],
            interface_name=self.route["interfaceName"],
            next_hop=self.route["nextHop"],
            next_hop_type=self.route["nextHopType"],
            vrf=self.route["vrf"],
        )
        assert newRoute is None

    @responses.activate
    def test_add_generic_routes_400(self, st):
        """Add one or multiple routes"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/route",
            status=400,
        )

        """Add single route"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_route(
                device=self.route["mgmtId"],
                destination=self.route["destination"],
                mask=self.route["mask"],
                interface_name=self.route["interfaceName"],
                next_hop=self.route["nextHop"],
                next_hop_type=self.route["nextHopType"],
                vrf=self.route["vrf"],
            )
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_add_generic_routes_404(self, st):
        """Add one or multiple routes"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/route",
            status=404,
        )

        """Add single route"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_route(
                device=self.route["mgmtId"],
                destination=self.route["destination"],
                mask=self.route["mask"],
                interface_name=self.route["interfaceName"],
                next_hop=self.route["nextHop"],
                next_hop_type=self.route["nextHopType"],
                vrf=self.route["vrf"],
            )
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_route(self, st, generic_route_mock):
        """Get route by route id"""
        mockRes = get_api_node(self.route27, "GenericRoute")
        route = GenericRoute.kwargify(mockRes)
        routeByInt = st.get_generic_route(27)
        assert routeByInt == route

        with pytest.raises(ValueError) as exception:
            st.get_generic_route(404)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_routes(self, st, generic_route_mock):
        """Get routes by mgmt id"""
        routes = [
            GenericRoute.kwargify(d)
            for d in get_api_node(self.mgmt, "GenericRoutes", listify=True)
        ]
        routesByInt = st.get_generic_routes(3)
        assert routesByInt == routes

        with pytest.raises(ValueError) as exception:
            st.get_generic_routes(404)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_update_generic_routes_200(self, st):
        """Update one or multiple routes"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/route",
            status=200,
        )

        """Update single route"""
        route_to_update = GenericRoute.kwargify(self.route)
        newRoute = st.update_generic_route(route_to_update)
        assert newRoute is None

    @responses.activate
    def test_update_generic_routes_400(self, st):
        """PUT bad request"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/route",
            status=400,
        )

        with pytest.raises(ValueError) as exception:
            route_to_update = GenericRoute.kwargify(self.route)
            st.update_generic_route(route_to_update)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_update_generic_routes_404(self, st):
        """PUT mgmt id not found"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/route",
            status=404,
        )

        with pytest.raises(ValueError) as exception:
            route_to_update = GenericRoute.kwargify(self.route)
            st.update_generic_route(route_to_update)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_route(self, st, generic_route_mock):
        """Delete route by route id"""
        routeByInt = st.delete_generic_route(27)
        assert routeByInt is None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_route(404)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_routes(self, st, generic_route_mock):
        """Delete routes by mgmt id"""
        routesByInt = st.delete_generic_routes(2)
        assert routesByInt is None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_routes(404)
        assert "Not Found" in str(exception.value)



from enum import Enum
from typing import Optional

from pytos2.utils import propify, prop, kwargify
from pytos2.models import Jsonable


@propify
class GenericDevice(Jsonable):
    id: Optional[int] = prop(None)
    customer_id: Optional[int] = prop(None)
    name: Optional[str] = prop(None)



from enum import Enum
from typing import Optional

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, stringify_optional_obj

from netaddr import IPAddress


@propify
class GenericRoute(Jsonable):
    id: Optional[int] = prop(None, converter=optional(int))
    device_id: Optional[int] = prop(None, converter=optional(int), key="mgmtId")
    destination: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )
    mask: Optional[IPAddress] = prop(
        None,
        converter=optional(IPAddress),
        jsonify=stringify_optional_obj,
    )
    interface_name: str = prop(None, key="interfaceName")
    next_hop: Optional[str] = prop(None, key="nextHop")
    next_hop_type: Optional[str] = prop(None, key="nextHopType")
    vrf: Optional[str] = prop(None)



import pytest
import responses

import json
import os
from datetime import datetime, date, time
from time import mktime
from requests.exceptions import HTTPError
from . import conftest

from pytos2.securetrack.revision import Revision, RevisionTicket
from pytos2.utils import get_api_node


class TestRevision:
    @pytest.fixture
    def all_revisions(self):
        json_dir = "tests/securetrack/json/revisions"

        files = os.listdir(json_dir)
        files = [
            name
            for name in files
            if name.startswith("device-") and name.endswith(".json")
        ]

        json_list = [json.load(open(os.path.join(json_dir, name))) for name in files]
        return json_list

    def get_revision(self, rev_num):
        j = json.load(
            open(f"tests/securetrack/json/revisions/device-{rev_num}.json", "r")
        )
        return get_api_node(j, "revision", listify=True)

    @pytest.fixture
    def revisions_1(self):
        return self.get_revision(1)

    @pytest.fixture
    def revisions_20(self):
        return self.get_revision(20)

    @pytest.fixture
    def revision_with_modules_and_policy(self, revisions_20):
        for r in revisions_20:
            if int(r["id"]) == 118:
                return Revision.kwargify(r)

    @pytest.fixture
    def revision_with_ticket(self, revisions_1):
        for r in revisions_1:
            if int(r["id"]) == 719:
                return Revision.kwargify(r)

        raise ValueError(
            "Revision 719 in device ID 1 revisions no longer has an attached ticket"
        )

    def test_list_attributes(
        self, revision_with_ticket, revision_with_modules_and_policy
    ):
        assert revision_with_ticket.tickets[0].source == RevisionTicket.Source.SCW

        assert len(revision_with_ticket.tickets) == 2
        assert revision_with_ticket.tickets[0].id == 22

        assert revision_with_modules_and_policy.modules_and_policy[0].module == "SMCPM"
        assert (
            revision_with_modules_and_policy.modules_and_policy[0].policy == "Standard"
        )

    def test_attributes(self, revisions_1):
        revision = revisions_1[0]
        revision = Revision.kwargify(revision)

        assert revision.id == 1674
        assert revision.revision_id == 39
        assert revision.action == Revision.Action.AUTOMATIC
        assert revision.date == date(2018, 3, 8)
        assert revision.time == time(12, 47, 17)
        assert revision.admin == "-"
        assert revision.gui_client == "-"
        assert revision.audit_log == "-"
        assert revision.policy_package == "Standard"
        assert revision.authorization_status == Revision.AuthorizationStatus.AUTHORIZED

        assert isinstance(revision.modules_and_policy, list)
        assert len(revision.modules_and_policy) == 0

        assert revision.firewall_status
        assert revision.is_ready

        assert isinstance(revision.tickets, list)
        assert len(revision.tickets) == 0

    def test_set_attributes(self, revisions_1):
        revision = Revision.kwargify(revisions_1[0])

        revision.date = "2015-03-02"
        revision.time = "14:14:50.300"

        assert revision._date == date(2015, 3, 2)
        assert revision._time == time(14, 14, 50, 300000)

        revision.date = date(2014, 4, 12)
        assert revision._date == date(2014, 4, 12)

        revision.date = datetime(2014, 4, 11, 12, 12, 50)
        assert type(revision._date) is date
        assert revision._date == date(2014, 4, 11)
        assert revision.date_str == "2014-04-11"

        with pytest.raises(ValueError) as context:
            revision.date = []

        revision.date = None
        assert revision._date is None

        revision.time = time(12, 30)
        assert revision._time == time(12, 30, 0)

        revision.time = datetime(2013, 3, 2, 12, 50)
        assert revision._time == time(12, 50)
        assert revision.time_str == "12:50:00.000"

        with pytest.raises(ValueError) as context:
            revision.time = []

        revision.time = None
        assert revision._time is None

        _t = datetime(2015, 5, 5, 12, 30)
        _timestamp = mktime(_t.timetuple())

        revision.unix_timestamp = _timestamp
        assert revision.date == date(2015, 5, 5)
        assert revision.time == time(12, 30)
        assert revision.unix_timestamp == _timestamp

        with pytest.raises(ValueError) as context:
            revision.unix_timestamp = "invalid"

    def test_all_kwargify(self, all_revisions):
        for i, _revisions in enumerate(all_revisions):
            if _revisions is None:
                continue

            revisions = get_api_node(_revisions, "revision", listify=True)
            for revision in revisions:
                Revision.kwargify(revision)

    @responses.activate
    def test_revision_error(self, st, revisions_mock):
        temp = st.cache
        st.cache = None  # force cache to be empty
        with pytest.raises(HTTPError):
            st.get_revisions(device=100000)
        st.cache = None
        with pytest.raises(HTTPError):
            st._get_revisions_from_server(100000)
        st.cache = temp



from typing import Optional, List
from enum import Enum

from pytos2.utils import propify, prop
from pytos2.models import Jsonable

from .rule import SlimRule


@propify
class MarkedObjects(Jsonable):
    uid: str = prop(None)


@propify
class IntersectingObjects(Jsonable):
    found_in_source: Optional[MarkedObjects] = prop(None)
    found_in_destination: Optional[MarkedObjects] = prop(None)
    found_in_service: Optional[MarkedObjects] = prop(None)


@propify
class RelatedRuleContainer(Jsonable):
    ignore: Optional[bool] = prop(None)
    rule: Optional[SlimRule] = prop(None)
    intersecting_objects: Optional[IntersectingObjects] = prop(None)


@propify
class RelatedRulesBinding(Jsonable):
    related_rules: List[RelatedRuleContainer] = prop(
        factory=list, flatify="related_rule", key="rules"
    )


@propify
class RelatedRulesDevice(Jsonable):
    bindings: List[RelatedRulesBinding] = prop(
        factory=list, flatify="binding", key="bindings"
    )
    management_id: Optional[int] = prop(None)


@propify
class RelatedRulesAccessRequest(Jsonable):
    ar: Optional[int] = prop(None, key="AR")
    devices: List[RelatedRulesDevice] = prop(
        factory=list, flatify="device", key="devices"
    )


@propify
class RelatedRulesResult(Jsonable):
    class Meta(Enum):
        ROOT = "related_rules"

    access_requests: List[RelatedRulesAccessRequest] = prop(
        factory=list, flatify="access_request", key="access_requests"
    )



from enum import Enum
from typing import Optional

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, stringify_optional_obj

from netaddr import IPAddress


@propify
class GenericInterface(Jsonable):
    class Type(Enum):
        EXTERNAL = "external"
        INTERNAL = "internal"

    id: Optional[int] = prop(None, converter=optional(int))
    device_id: Optional[int] = prop(None, converter=optional(int), key="mgmtId")
    name: str = prop(None)
    ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )
    mask: Optional[IPAddress] = prop(
        None,
        converter=optional(IPAddress),
        jsonify=stringify_optional_obj,
    )
    vrf: Optional[str] = prop(None)
    mpls: Optional[bool] = prop(False)
    unnumbered: Optional[bool] = prop(False)
    type: Optional[Type] = prop(None)



from enum import Enum
from typing import Optional, List, Union

from attr.converters import optional
from netaddr import IPAddress, IPNetwork, IPRange  # type: ignore
from netaddr.core import AddrFormatError

from pytos2.models import FQDNIp, Jsonable, UnMapped, IPType
from pytos2.utils import propify, prop, kwargify, get_api_node


class NatXsiType(Enum):
    NAT_INFO = "natInfoDTO"
    MIP_NAT_INFO = "mipNatInfoDTO"
    VIP_NAT_INFO = "vipNatInfoDTO"
    DIP_NAT_INFO = "dipNatInfoDTO"
    CHECKPOINT_NAT_INFO = "checkpointNatInfoDTO"
    FORTIGATE_NAT_INFO = "fortigateNatInfoDTO"


class NetworkObjectXsiType(Enum):
    BASIC_NETWORK_OBJECT = "basicNetworkObjectDTO"
    CLOUD_SECURITY_GROUP = "cloudSecurityGroupDTO"
    DOMAIN_NETWORK_OBJECT = "DomainNetworkObjectDTO"
    HOST_NETWORK_OBJECT = "hostNetworkObjectDTO"
    HOST_NETWORK_OBJECT_WITH_INTERFACES = "hostNetworkObjectWithInterfacesDTO"
    IDENTITY_AWARENESS = "identityAwarenessDTO"
    IDENTITY_AWARENESS_USER = "identityAwarenessUserDTO"
    NETWORK_OBJECT = "networkObjectDTO"
    NETWORK_OBJECT_GROUP = "networkObjectGroupDTO"
    NETWORK_OBJECT_VIRTUAL_SERVER = "networkObjectVirtualServerDTO"
    RANGE_NETWORK_OBJECT = "rangeNetworkObjectDTO"
    SUBNET_NETWORK_OBJECT = "subnetNetworkObjectDTO"
    VM_NETWORK_OBJECT = "VMInstanceDTO"


@propify
class NatInfo(Jsonable):
    xsi_type: Optional[NatXsiType] = prop(
        None, key=Jsonable.Prop.XSI_TYPE.value, repr=False
    )
    interface_name: Optional[str] = prop(None)


@propify
class MipNatInfo(NatInfo):
    virtual_router: Optional[str] = prop(None)
    mapped_to_ip: Optional[str] = prop(None)


@propify
class DipNatInfo(NatInfo):
    shift_from: Optional[str] = prop(None)


@propify
class CheckpointNatInfo(NatInfo):
    class Prop(Enum):
        MAPPED_TO_IP = "mapped_to_ip"

    class CheckpointNatType(Enum):
        STATIC = "STATIC"
        HIDDEN = "HIDDEN"

    checkpoint_nat_type: Optional[CheckpointNatType] = prop(None)
    mapped_to_ip: Optional[Union[str, IPAddress]] = prop(None, kwargify=False)

    @classmethod
    def kwargify(cls, obj):
        _obj, kwargs = kwargify(cls, obj)
        mapped_to_ip = _obj.get(cls.Prop.MAPPED_TO_IP.value)
        if mapped_to_ip:
            try:
                kwargs["mapped_to_ip"] = IPAddress(mapped_to_ip)
            except AddrFormatError:
                pass

        return cls(**kwargs)


@propify
class FortigateNatInfo(NatInfo):
    class Prop(Enum):
        MAPPED_IP = "mapped-ip"
        MAPPED_IP_MAX = "mapped-ip-max"
        EXTERNAL_PORT = "external-port"
        MAPPED_PORT = "mapped-port"

    port_forwarding_type: Optional[str] = prop(None)
    src_filter: Optional[str] = prop(None)
    src_interface_filter: Optional[str] = prop(None)
    mapped_ip: Optional[IPAddress] = prop(None, key=Prop.MAPPED_IP.value)
    mapped_ip_max: Optional[IPAddress] = prop(None, key=Prop.MAPPED_IP_MAX.value)
    external_port: Optional[str] = prop(None, key=Prop.EXTERNAL_PORT.value)
    mapped_port: Optional[str] = prop(None, key=Prop.MAPPED_PORT.value)
    forti_vip: Optional[bool] = prop(None)
    Interface: Optional[str] = prop(None)


def classify_nat_type(obj):
    return (
        {
            NatXsiType.CHECKPOINT_NAT_INFO.value: CheckpointNatInfo,
            NatXsiType.DIP_NAT_INFO.value: DipNatInfo,
            NatXsiType.FORTIGATE_NAT_INFO.value: FortigateNatInfo,
            NatXsiType.MIP_NAT_INFO.value: MipNatInfo,
            NatXsiType.VIP_NAT_INFO.value: NatInfo,
        }
        .get(obj.get(Jsonable.Prop.XSI_TYPE.value), UnMapped)
        .kwargify(obj)
    )


NatType = Union[CheckpointNatInfo, DipNatInfo, FortigateNatInfo, MipNatInfo, NatInfo]


@propify
class NetworkObject(Jsonable):
    IPType = IPType

    class Type(Enum):
        BASIC = "basic"
        HOST = "host"
        RANGE = "range"
        SUBNET = "subnet"
        GROUP = "group"
        DOMAIN = "domain"
        ANY = "any"
        INSTALLON = "installon"
        HOST_WITH_IFACES = "hostWithIfaces"
        USER = "user"
        USER_GROUP = "user-group"
        ALIAS = "alias"
        SECURITY_GROUP = "security_group"
        VM_INSTANCE = "vm_instance"
        INTERNET = "internet"
        REFERENCE = "reference"
        IDENTITY_AWARENESS = "identity_awareness"
        DOMAIN_NETWORK = "domain-network"

    class ClassName(Enum):
        ADDRESS_RANGE = "address_range"
        MULTICAST_ADDRESS_RANGE = "multicast_address_range"
        VPN_ROUTE = "vpn_route"
        INTERFACE = "interface"
        NETWORK_INTERFACE = "network_interface"
        FIREWALL_POLICY = "firewall_policy"
        SUB_POLICY = "sub_policy"
        SECURITY_RULE = "security_rule"
        ROUTING_ENTRY = "routing_entry"
        SECURITY_RULES_GROUP = "security_rules_group"
        RULE_SOURCE = "rule_source"
        RULE_DESTINATION = "rule_destination"
        RULE_INSTALL = "rule_install"
        OTHER_SERVICE = "other_service"
        OTHER_OBJECT = "other_object"
        ACCEPT_ACTION = "accept_action"
        DROP_ACTION = "drop_action"
        DENY_ACTION = "deny_action"
        CONTINUE_ACTION = "continue_action"
        TRUST_ACTION = "trust_action"
        RESET_CLIENT_ACTION = "reset_client_action"
        RESET_SERVER_ACTION = "reset_server_action"
        RESET_BOTH = "reset_both_action"
        NETWORK_OBJECT_GROUP = "network_object_group"
        GROUP_WITH_EXCEPTION = "group_with_exception"
        NETWORK = "network"
        HOST = "host_plain"
        FQDN = "fqdn"
        VIRTUAL_MACHINE_INSTANCE = "vm_instance"
        CONTROLLER = "controllers"
        TCP_SERVICE = "tcp_service"
        UDP_SERVICE = "udp_service"
        SCTP_SERVICE = "sctp_service"
        ICMP_SERVICE = "icmp_service"
        INSTALL_ON = "install_on"
        IP_SERVICE = "ip_service"
        SECURITY_HEADER_RULE = "security_header_rule"
        ZONE = "zone"
        DOMAIN = "domain"
        INTERNET_SERVICE = "internet-service"
        IPV6 = "ipv6_object"
        Range = "address_range"
        SERVICE_GROUP = "service_group"
        POLICIES_COLLECTION = "policies_collection"
        REJECT_ACTION = "reject_action"
        USER_AUTHENTICATE = "user_authenticate"
        CLIENT_AUTHENTICATE = "client_authenticate"
        UNSUPPORTED_NETWORK_OBJECT = "unsupported_object"
        SCHEDULED_EVENT = "scheduled_event"
        TIME = "time"
        TIME_PERIOD = "time_period"
        NAT_POOL = "nat_pool"
        NAT_HEADER = "nat_header_rule"
        ADTR_TRANSLATION_RULE = "address_translation_rule"
        TRANSLATE_STATIC = "translate_static"
        TRANSLATE_HIDDEN = "translate_hidden"
        TRANSLATE_SERVICE = "service_translate"
        BEHIND_INTERFACE_NAT = "behind_interface_nat"
        SECURITY_PROFILE = "security_profile"
        SECURITY_PROFILE_GROUP = "security_profile_group"
        RULE_SERVICES = "rule_services"
        RULE_VPN = "rule_vpn"
        ICMPV6_SERVICE = "icmpv6_service"
        ANY_OBJECT = "any_object"
        ACCESS_ROLE = "access_role"
        USER = "user"
        USER_GROUP = "user_group"
        AD_BRANCH = "ad_branch"
        AD_GROUP = "ad_group"
        AD_USER = "ad_user"
        ALL_IDENTIFIED = "all_identified"
        ENCRYPT = "encrypt"
        CLIENT_ENCRYPT = "client_encrypt"
        RPC_SERVICE = "rpc_service"
        DCE_RPC_SERVICE = "dcerpc_service"
        UNSUPPORTED_SERVICE = "unsupported_service"
        INLINE_LAYER_ACTION = "inline_layer_action"
        SUB_POLICY_COLLECTION = "sub_policy_collection"
        DYNAMIC_OBJECT = "dynamic_object"
        HOST_CKP = "host_ckp"
        GATEWAY_CKP = "gateway_ckp"
        SOFTWARE_GATEWAY = "sofaware_gateway"
        EMBEDDED_DEVICE = "embedded_device"
        GATEWAY_CLUSTER = "gateway_cluster"
        VS_CLUSTER_NETOBJ = "vs_cluster_netobj"
        VSX_CLUSTER_NETOBJ = "vsx_cluster_netobj"
        CLUSTER_MEMBER = "cluster_member"
        GPRS_APN = "gprs_apn"
        UTM_CLUSTER_MEMBER = "utm_cluster_member"
        VS_CLUSTER_MEMBER = "vs_cluster_member"
        VSX_CLUSTER_MEMBER = "vsx_cluster_member"
        VS_NETOBJ = "vs_netobj"
        VSX_NETOBJ = "vsx_netobj"
        SECURITY_ZONE = "security_zone"
        GATEWAY_PLAIN = "gateway_plain"
        GSN_HANDOVER_GROUP = "gsn_handover_group"
        DATA_CENTER = "data_center"
        DATA_CENTER_OBJECT = "data_center_object"
        ENDPOINT_GROUP = "endpoint_group"
        BRIDGE_DOMAIN = "bridge_domain"
        LEARNED_ENDPOINT = "learned_endpoint"
        EXTERNAL_ROUTED_NETWORK = "external_routed_network"
        LOGICAL_NODE_PROFILE = "logical_node_profile"
        LOGICAL_INTERFACE_PROFILE = "logical_interface_profile"
        EXTERNAL_EPG = "external_endpoint_group"
        CONFIGURED_NODE = "configured_node"
        EXTERNAL_INTERFACE = "external_interface"
        EPGS_GROUP = "epgs_group"

    class Prop(Enum):
        GLOBAL = "global"

    xsi_type: Optional[NetworkObjectXsiType] = prop(
        None, key=Jsonable.Prop.XSI_TYPE.value, repr=False
    )
    type_on_device: Optional[str] = prop(None, repr=False)
    domain_name: Optional[str] = prop(None, repr=False)
    ip_type: Optional[IPType] = prop(None, repr=False)
    any_zone: bool = prop(False, repr=False)
    management_domain_securetrack_name: Optional[str] = prop(None, repr=False)
    nat_info: Optional[NatType] = prop(None, repr=False, kwargify=classify_nat_type)
    implicit: bool = prop(False, repr=False)
    management_domain: Optional[str] = prop(None, repr=False)
    overrides: bool = prop(False, repr=False)
    type: Type = prop(Type.BASIC, repr=False)
    deviceName: Optional[str] = prop(None, repr=False)
    device_id: Optional[int] = prop(None, converter=optional(int), repr=False)
    application_id: Optional[int] = prop(None, converter=optional(int), repr=False)
    uid: Optional[str] = prop(None, repr=False)
    comment: Optional[str] = prop(None, repr=False)
    is_global: bool = prop(False, key=Prop.GLOBAL.value, repr=False)
    name: Optional[str] = prop(None)
    class_name: Optional[ClassName] = prop(None, repr=False)
    display_name: Optional[str] = prop(None)


@propify
class BasicNetworkObject(NetworkObject):
    pass


FQDNIP = FQDNIp


@propify
class DomainNetworkObject(NetworkObject):
    class Prop(Enum):
        IPS = "ips"

    domain: Optional[str] = prop(None, repr=False)
    ips: List[FQDNIP] = prop(factory=list, flatify="ip")


@propify
class PoolMember(Jsonable):
    class Prop(Enum):
        IP = "ip"
        NETMASK = "netmask"

    netmask: Optional[IPAddress] = prop(None)
    ip: Optional[IPAddress] = prop(None)
    name: Optional[str] = prop(None)


@propify
class NetworkObjectVirtualServer(NetworkObject):
    class Prop(Enum):
        VIRTUAL_IP = "virtual_ip"
        NETMASK = "netmask"
        POOL_MEMBER = "pool_member"

    class Protocol(Enum):
        TCP = "TCP"
        UDP = "UDP"
        SCTP = "SCTP"

    netmask: Optional[IPAddress] = prop(None)
    f5_device_name: Optional[str] = prop(None)
    virtual_ip: Optional[IPAddress] = prop(None)
    pool_member: List[PoolMember] = prop(
        factory=list, repr=False, key=Prop.POOL_MEMBER.value
    )
    protocol: Optional[Protocol] = prop(None)
    port: Optional[str] = prop(None)


@propify
class HostNetworkObject(NetworkObject):
    class Prop(Enum):
        IP = "ip"

    xsi_type: NetworkObjectXsiType = prop(NetworkObjectXsiType.HOST_NETWORK_OBJECT)
    ip: Optional[IPAddress] = prop(None)


@propify
class InterfaceIP(Jsonable):
    class Visibility(Enum):
        PRIVATE = "private"
        PUBLIC = "public"

    class Precedence(Enum):
        PRIMARY = "primary"
        SECONDARY = "secondary"

    netmask: Optional[IPAddress] = prop(None)
    visibility: Optional[Visibility] = prop(None)
    precedence: Optional[Precedence] = prop(None)
    vendorAttachmentPolicy: Optional[str] = prop(None)
    ip: Optional[IPAddress] = prop(None)


@propify
class Interface(Jsonable):
    name: Optional[str] = prop(None)
    interface_ips: List[InterfaceIP] = prop(factory=list, flatify="interface_ip")


@propify
class HostNetworkObjectWithInterfaces(HostNetworkObject):
    interfaces: List[Interface] = prop(factory=list, flatify="interface")


@propify
class SubnetNetworkObject(NetworkObject):
    class Prop(Enum):
        IP = "ip"
        NETMASK = "netmask"

    subnet: Optional[IPNetwork] = None

    @classmethod
    def kwargify(cls, obj):
        _obj, kwargs = kwargify(cls, obj)

        try:
            if cls.Prop.IP.value in _obj and cls.Prop.NETMASK.value in _obj:
                kwargs["subnet"] = IPNetwork(
                    f"{_obj[cls.Prop.IP.value]}/{_obj[cls.Prop.NETMASK.value]}"
                )
        except AddrFormatError:
            pass

        kwargs["type"] = kwargs.get("type", NetworkObject.Type.SUBNET)

        return cls(**kwargs)


@propify
class RangeNetworkObject(NetworkObject):
    class Prop(Enum):
        FIRST_IP = "first_ip"
        LAST_IP = "last_ip"

    range: Optional[IPRange] = prop(None)

    """
    reverse_range: Set to true if first is greater than last on the
    source network object.
    """
    reverse_range: bool = prop(False)

    @classmethod
    def kwargify(cls, obj):
        _obj, kwargs = kwargify(cls, obj)

        kwargs["type"] = kwargs.get("type", NetworkObject.Type.RANGE)

        first = _obj.get(cls.Prop.FIRST_IP.value)
        last = _obj.get(cls.Prop.LAST_IP.value)

        try:
            start = IPAddress(first)
            end = IPAddress(last)
            kwargs["reverse_range"] = start > end
            kwargs["range"] = IPRange(min(start, end), max(start, end))
        except (AddrFormatError, ValueError) as e:
            msg = f"First IP {first} / Last IP {last} could not be cast to IPAddress / IPRange"
            raise ValueError(msg) from e

        return cls(**kwargs)

    @property
    def first(self):
        return self.range.first if not self.reverse_range else self.range.last

    @property
    def last(self):
        return self.range.last if not self.reverse_range else self.range.first


@propify
class ObjectReference(Jsonable):
    ip: Optional[IPAddress] = prop(None)
    uid: Optional[str] = prop(None)
    name: Optional[str] = prop(None)
    members: List["ObjectReference"] = prop(factory=list, flatify="member")
    type: Optional[NetworkObject.Type] = prop(None)
    display_name: Optional[str] = prop(None)

    @classmethod
    def kwargify(cls, obj):
        _obj, kwargs = kwargify(cls, obj)
        # edge case of some strange rules on ASAs
        if "DM_INLINE_members" in _obj:
            kwargs["members"] = [
                cls.kwargify(o) for o in get_api_node(obj, "DM_INLINE_members.member")
            ]
        return cls(**kwargs)


@propify
class NetworkObjectGroup(NetworkObject):
    class Prop(Enum):
        EXCLUSION = "exclusion"
        MEMBER = "member"

    access_allowed: Optional[bool] = prop(None)
    exclusions: List[ObjectReference] = prop(
        factory=list, key=Prop.EXCLUSION.value, repr=False
    )
    members: List[ObjectReference] = prop(
        factory=list, key=Prop.MEMBER.value, repr=False
    )


@propify
class CloudSecurityGroup(NetworkObjectGroup):
    pass


@propify
class IdentityAwarenessUser(Jsonable):
    class Prop(Enum):
        GLOBAL = "global"

    xsi_type: Optional[str] = prop(None, key=Jsonable.Prop.XSI_TYPE.value, repr=False)
    dn: Optional[str] = prop(None)
    type: Optional[NetworkObject.Type] = prop(None)
    deviceName: Optional[str] = prop(None)
    device_id: Optional[int] = prop(None, converter=optional(int))
    application_id: Optional[int] = prop(None, converter=optional(int))
    uid: Optional[str] = prop(None)
    comment: Optional[str] = prop(None)
    is_global: bool = prop(False, key=Prop.GLOBAL.value)
    name: Optional[str] = prop(False)
    class_name: Optional[NetworkObject.ClassName] = prop(None)
    display_name: Optional[str] = prop(None)


@propify
class IdentityAwareness(NetworkObject):
    class Prop(Enum):
        USERS = "users"
        NETWORKS = "networks"

    users: List[IdentityAwarenessUser] = prop(factory=list)
    networks: List[ObjectReference] = prop(factory=list)


@propify
class VMNetworkObject(NetworkObject):
    xsi_type: NetworkObjectXsiType = prop(NetworkObjectXsiType.VM_NETWORK_OBJECT)
    id: Optional[str] = prop(None)
    type: NetworkObject.Type = prop(NetworkObject.Type.HOST)
    ip: Optional[IPAddress] = prop(None)
    interfaces: List[Interface] = prop(factory=list, flatify="interface")
    state: Optional[str] = prop(None)


def classify_network_object(obj):
    return (
        {
            NetworkObjectXsiType.BASIC_NETWORK_OBJECT.value: BasicNetworkObject,
            NetworkObjectXsiType.CLOUD_SECURITY_GROUP.value: CloudSecurityGroup,
            NetworkObjectXsiType.DOMAIN_NETWORK_OBJECT.value: DomainNetworkObject,
            NetworkObjectXsiType.HOST_NETWORK_OBJECT.value: HostNetworkObject,
            NetworkObjectXsiType.HOST_NETWORK_OBJECT_WITH_INTERFACES.value: HostNetworkObjectWithInterfaces,
            NetworkObjectXsiType.IDENTITY_AWARENESS.value: IdentityAwareness,
            NetworkObjectXsiType.NETWORK_OBJECT.value: NetworkObject,
            NetworkObjectXsiType.NETWORK_OBJECT_GROUP.value: NetworkObjectGroup,
            NetworkObjectXsiType.NETWORK_OBJECT_VIRTUAL_SERVER.value: NetworkObjectVirtualServer,
            NetworkObjectXsiType.RANGE_NETWORK_OBJECT.value: RangeNetworkObject,
            NetworkObjectXsiType.SUBNET_NETWORK_OBJECT.value: SubnetNetworkObject,
            NetworkObjectXsiType.VM_NETWORK_OBJECT.value: VMNetworkObject,
        }
        .get(obj.get(Jsonable.Prop.XSI_TYPE.value), UnMapped)
        .kwargify(obj)
    )



import pytest
from pytos2.utils.cache import Cache, CacheIndex


class Obj:
    def __init__(self, a, b):
        self.a = a
        self.b = b


class TestCache:
    def test_destruct(self):
        cache = Cache()
        index = cache.make_index("a")

        index.__del__()

    def test_add(self):
        cache = Cache()
        a_index = cache.make_index("a")
        b_index = cache.make_index("b")

        cache.add(Obj(5, 10))
        cache.add(Obj(20, 20))

        assert a_index.get(5).b == 10
        assert b_index.get(10).a == 5
        assert a_index.get(20).b == 20

        cache.add(Obj(5, 20))
        assert a_index.get(5).b == 10
        assert a_index.get(5, 1).b == 20

    def test_is_empty(self):
        cache = Cache()
        index = cache.make_index("a")

        assert cache.is_empty()

        o = Obj(5, 10)
        cache.add(o)

        assert not cache.is_empty()

        cache.remove(o)

        assert cache.is_empty()

    def test_remove(self):
        cache = Cache()
        index = cache.make_index("a")

        o1 = Obj(5, 10)
        o2 = Obj(20, 20)

        cache.add(o1)
        cache.add(o2)

        assert index.get(5).b == 10
        assert index.get(5).a == 5

        cache.remove(o1)
        assert index.get(5) is None

    def test_clear(self):
        cache = Cache()
        index = cache.make_index("a")

        o1 = Obj(5, 10)
        o2 = Obj(20, 20)

        cache.add(o1)
        cache.add(o2)

        assert index.get(5).b == 10
        assert index.get(20).a == 20

        cache.clear()

        assert cache.get_data() == []
        assert cache.is_empty()

    def test_set(self):
        cache = Cache()
        index = cache.make_index("a")

        arr = [Obj(5, 10), Obj(20, 20)]

        cache.set_data(arr)

        assert index.get(5).b == 10
        assert index.get(20).a == 20



from enum import Enum
from typing import Optional

from pytos2.utils import propify, prop
from pytos2.models import Jsonable


@propify
class Binding(Jsonable):
    class XsiType(Enum):
        POLICY_BINDING = "policy__binding"

    xsi_type: XsiType = prop(XsiType.POLICY_BINDING, key=Jsonable.Prop.XSI_TYPE.value)

    policy_name: Optional[str] = prop(None)
    installed_on_module: Optional[str] = prop(None)



import pytest
import json

import responses

from netaddr import IPAddress

from typing import List
from pytos2.securetrack.device import Device, DeviceLicenseStatus
from pytos2.securetrack.managed_devices import (
    BulkOperationTask,
    BulkOperationTaskResult,
    BulkOperationTaskResultDevice,
    BulkOperationTaskStatusList,
)
from pytos2.utils import get_api_node


class TestDevice:
    @pytest.fixture
    def device(self):
        j = json.load(open("tests/securetrack/json/devices/devices.json"))
        device_node = get_api_node(j, "devices.device")[0]
        return Device.kwargify(device_node)

    def test_attributes(self, device):
        assert device.id == 1
        assert device.name == "RTR1"
        assert device.vendor == Device.Vendor.CISCO
        assert device.model == Device.Model.ROUTER
        assert isinstance(device.domain_id, int) and device.domain_id == 1
        assert device.domain_name == "Default"
        assert device.module_uid == ""
        assert device.module_type == "IOS"
        assert device.status == "Stopped"
        assert device.ip == IPAddress("10.100.200.54")
        assert (
            isinstance(device.latest_revision, int) and device.latest_revision == 1674
        )
        assert device.virtual_type == ""
        assert len(device.licenses) == 3
        assert isinstance(device.licenses[0], DeviceLicenseStatus)

    def test_set_attributes(self, device):
        assert device.ip == IPAddress("10.100.200.54")

        device.ip = "1.2.3.4"
        assert device.ip == IPAddress("1.2.3.4")

    @responses.activate
    def test_properties(self, devices_mock, st):
        device = st.get_device(identifier=60)
        assert device
        assert device.name == "NSX-Edge-01"

        parent_device = device.parent
        assert parent_device.id == 58
        assert parent_device.name == "NSX"

        assert len(parent_device.children) == 4

        grandparent = parent_device.parent
        assert grandparent is None

    @responses.activate
    def test_bulk_delete_devices(self, devices_mock, st):
        task: BulkOperationTask = st.bulk_delete_devices([5, 8])
        assert isinstance(task.uid, str)

    @responses.activate
    def test_bulk_update_topology(self, devices_mock, st):
        task: BulkOperationTask = st.bulk_update_topology([5, 8])
        assert isinstance(task.uid, str)

    @responses.activate
    def test_get_bulk_device_task(self, devices_mock, st):
        task_result: BulkOperationTaskResult = st.get_bulk_device_task(
            "ada853b8-46f7-474b-bb4e-3309a3a9d0af"
        )
        assert isinstance(task_result.in_progress, BulkOperationTaskStatusList)
        assert all(
            isinstance(item, BulkOperationTaskResultDevice)
            for item in task_result.in_progress.devices
        )

        device = task_result.in_progress.devices[0]
        assert device.id == 1
        assert device.name == "Europe-CMA"
        assert device.ip_address == "192.168.1.1"
        assert device.status is None

        assert isinstance(task_result.succeeded, BulkOperationTaskStatusList)
        assert all(
            isinstance(item, BulkOperationTaskResultDevice)
            for item in task_result.succeeded.devices
        )

        assert isinstance(task_result.failed, BulkOperationTaskStatusList)
        assert all(
            isinstance(item, BulkOperationTaskResultDevice)
            for item in task_result.failed.devices
        )

        assert task_result.total_in_progress == 1
        assert task_result.total_failed == 0
        assert task_result.total_succeeded == 1

        # Check that status.description is used as status properly.
        task_result: BulkOperationTaskResult = st.get_bulk_device_task(
            "c878cb8c-6a6d-4939-b20b-550def656ac4"
        )
        device = task_result.failed.devices[0]
        assert device.status == "Device with id: 998 was not found"



import pytest
import json
import responses
from netaddr import IPAddress

from pytos2.securetrack.domain import Domain
from pytos2.utils import get_api_node


class TestDomains:
    @responses.activate
    def test_get_domain(self, domains_mock, st):
        domain = st.get_domain(cache=False, identifier=7)
        assert isinstance(domain, Domain)
        assert domain.id == 7
        assert domain.name == "London"
        domain = st.get_domain(cache=False, identifier="London")
        assert isinstance(domain, Domain)
        assert domain.id == 7
        assert domain.name == "London"
        domain_cache = st.get_domain(cache=True, identifier=7)
        assert isinstance(domain_cache, Domain)
        assert st._domains_cache.is_empty() is False
        domain_cache = st.get_domain(cache=True, identifier="London")
        assert isinstance(domain_cache, Domain)
        assert st._domains_cache.is_empty() is False

    @responses.activate
    def test_get_domains(self, domains_mock, st):
        domains = st.get_domains(cache=False)
        assert isinstance(domains, list)
        assert st._domains_cache.is_empty()
        domains_cache = st.get_domains(cache=True)
        assert isinstance(domains_cache, list)
        assert st._domains_cache.is_empty() is False

    @responses.activate
    def test_add_domain(self, domains_mock, st):
        res = st.add_domain("test", "test", "testaddr")
        assert isinstance(res, Domain)
        assert isinstance(res.id, int) and res.id == 7
        assert res.name == "London"

    @responses.activate
    def test_update_domain(self, domains_mock, st):
        res = st.update_domain(7, "test", "test", "testaddr")
        assert isinstance(res, Domain)



from enum import Enum
from typing import List, Optional, Union
from attr.converters import optional

from ..utils import propify, prop
from ..models import Jsonable


@propify
class BulkOperationTaskResultDevice(Jsonable):
    id: int = prop(key="device_id", converter=int)
    status: Optional[str] = prop(
        None, key="status", kwargify=lambda s: s.get("description", None) if s else None
    )
    name: Optional[str] = prop("", key="display_name")
    ip_address: Optional[str] = prop(None)


@propify
class BulkOperationTaskStatusList(Jsonable):
    count: int = prop(converter=int)
    devices: List[BulkOperationTaskResultDevice] = prop(flatify="devices", factory=list)


@propify
class BulkOperationTaskResult(Jsonable):
    in_progress: BulkOperationTaskStatusList = prop(key="in_progress", factory=list)
    succeeded: BulkOperationTaskStatusList = prop(key="succeeded", factory=list)
    failed: BulkOperationTaskStatusList = prop(key="failed", factory=list)
    total_in_progress: int = prop(0, converter=int)
    total_succeeded: int = prop(0, converter=int)
    total_failed: int = prop(0, converter=int)

    @property
    def count(self):
        return self.total_in_progress + self.total_succeeded + self.total_failed


@propify
class BulkOperationTask(Jsonable):
    uid: str = prop(key="task_uid")

    def get_result(self) -> BulkOperationTaskResult:
        from pytos2.securetrack.entrypoint import St

        if self.task_result:
            return self.task_result

        bulk_operation_task_result = St.default.get_bulk_device_task(self.uid)
        return bulk_operation_task_result.result



from enum import Enum
from typing import Optional, List

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, stringify_optional_obj

from netaddr import IPAddress


@propify
class RestCloudMember(Jsonable):
    name: str = prop(None)
    ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )


@propify
class RestAnonymousSubnet(Jsonable):
    ip: Optional[IPAddress] = prop(None)
    mask: Optional[IPAddress] = prop(None)


@propify
class RestCloud(Jsonable):
    class TopologyCloudType(Enum):
        NON_JOINED = "NON_JOINED"
        JOINED = "JOINED"
        MEMBER = "MEMBER"

    name: str = prop(None)
    domain: int = prop(converter=int)
    type: TopologyCloudType = prop(None)
    ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )
    members: List[RestCloudMember] = prop(factory=list)


@propify
class JoinCloud(Jsonable):
    name: str = prop(None)
    clouds: List[int] = prop(factory=list)


@propify
class SuggestedCloud(Jsonable):
    ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=stringify_optional_obj
    )
    management_name: str = prop("")
    management_id: Optional[int] = prop(None)
    routes_count: int = prop()
    cloud_name: str = prop("")
    is_parent: bool = prop()
    cloud_id: int = prop()
    parent_vertex_id: Optional[int] = prop(None)
    parent_cloud_name: Optional[str] = prop("")
    parent_cloud_id: Optional[int] = prop(None)
    vertex_id: int = prop()



import os

from setuptools import setup, find_namespace_packages


with open(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "README.md")
) as file:
    long_description = file.read()

setup(
    name="pytos2-ce",
    long_description=long_description,
    long_description_content_type="text/markdown",
    description="Pytos2 for tos1 tos2 and beyond",
    packages=find_namespace_packages(include=["pytos2*"]),
    use_scm_version=True,
    setup_requires=["setuptools_scm<7"],
    install_requires=[
        "requests<3",
        "traversify",
        "attrs>=19.2,<20",
        "python-json-logger<0.2",
        "netaddr<=0.10",
        "typing_extensions",
        "python-dateutil>=2.8,<3",
        "requests-oauthlib",
    ],
    extras_require={
        "dev": [
            "pytest",
            "pytest-cov",
            "pytest-logger",
            "responses",
            "black",
            "tox",
            "ruff",
        ],
        "dev-jupyter": ["black[jupyter]", "pre-commit", "nb-clean"],
    },
    test_suite="tests",
    package_data={"pytos2": ["py.typed"]},
)



import json
from pathlib import Path

from pytos2.api import Pager
import responses
import pytest
from netaddr import IPAddress, IPNetwork, IPRange

from pytos2.securetrack.network_object import (
    RangeNetworkObject,
    classify_network_object,
)
from pytos2.utils import get_api_node


class TestDevice:
    @pytest.fixture
    def device(self):
        j = json.load(open("tests/securetrack/json/devices/devices.json"))
        device_node = get_api_node(j, "devices.device")[0]
        return classify_network_object(device_node)


class TestLoadNetworkObjects:
    def test_all_objects(self):
        for path in Path("tests/securetrack/json/network_objects/").glob("*.json"):
            network_objects_node = get_api_node(
                json.load(path.open()), "network_objects.network_object"
            )
            if network_objects_node:
                for network_object in network_objects_node:
                    classify_network_object(network_object)


class TestCheckpointObject:
    @responses.activate
    def test_nat_object(self, st, network_objects_mock):
        # this should be refactored when get_network_object returns an object
        obj = st.get_network_object(
            device=20, name="CP_default_Office_Mode_addresses_pool"
        )
        assert obj.nat_info.mapped_to_ip == "Hide Behind Gateway"
        obj = st.get_network_object(device=20, name="Net_172.16.40.0")
        assert isinstance(obj.nat_info.mapped_to_ip, IPAddress)


class TestFortigateObject:
    @responses.activate
    def test_nat_object(self, st, network_objects_mock):
        # this should be refactored when get_network_object returns an object
        obj = st.get_network_object(device=157, name="VIP to HQ")
        assert isinstance(obj.nat_info.mapped_ip, IPAddress)
        assert isinstance(obj.nat_info.mapped_ip_max, IPAddress)


class TestHostObject:
    # this class covers host object with interfaces as well
    @responses.activate
    def test_host_object(self, st, network_objects_mock):
        # this should be refactored when get_network_object returns an object
        obj = st.get_network_object(device=184, name="SGW_200.237")
        assert isinstance(obj.ip, IPAddress)
        assert isinstance(obj.interfaces[0].interface_ips[0].ip, IPAddress)


class TestSubnetObject:
    @responses.activate
    def test_subnet_object(self, st, network_objects_mock):
        # this should be refactored when get_network_object returns an object
        obj = st.get_network_object(device=184, name="Sales_1")
        assert isinstance(obj.subnet, IPNetwork)


class TestRangeObject:
    ip_one = IPAddress("172.16.200.0")
    ip_two = IPAddress("172.16.200.255")
    ip_three = IPAddress("78.23.45.200")
    ip_four = IPAddress("78.32.45.10")

    @responses.activate
    def test_range_object(self, st, network_objects_mock):
        # this should be refactored when get_network_object returns an object
        obj = st.get_network_object(device=184, name="my_range")
        first = int(self.ip_one)
        last = int(self.ip_two)
        assert isinstance(obj.range, IPRange)
        assert obj.reverse_range is False
        assert obj.range.first == first
        assert obj.range.last == last
        assert obj.first == first
        assert obj.last == last

    @responses.activate
    def test_fortinet_range_object(self, st, network_objects_mock):
        obj = st.get_network_object(device=184, name="fortinet_reverse_range")
        first = int(self.ip_one)
        last = int(self.ip_two)
        assert isinstance(obj.range, IPRange)
        assert obj.range.first == first
        assert obj.range.last == last
        assert obj.reverse_range
        assert obj.first == last
        assert obj.last == first

    @responses.activate
    def test_real_worl_fortinet_range_object(self, st, network_objects_mock):
        obj = st.get_network_object(device=184, name="fortinet_real_world")
        first = int(self.ip_three)
        last = int(self.ip_four)
        assert isinstance(obj.range, IPRange)
        assert obj.range.first == first
        assert obj.range.last == last
        assert obj.reverse_range
        assert obj.first == last
        assert obj.last == first

    @responses.activate
    def test_classify_error(self):
        bad_obj = {"first_ip": "Not IP", "last_ip": "172.16.200.0"}

        with pytest.raises(ValueError):
            RangeNetworkObject.kwargify(bad_obj)


class TestNetworkObjects:
    @responses.activate
    def test_network_objects(self, st, network_objects_mock):
        pager = st.get_network_objects(revision=24522)
        assert isinstance(pager, Pager)
        network_objects = pager.fetch_all()
        obj = network_objects[0]
        assert obj.id == 450118
        assert obj.name == "Admin_server_02"

        network_objects = st.get_network_objects(revision=24522, object_ids=[450118])
        assert isinstance(network_objects, list)
        assert len(network_objects) == 1

        network_objects = st.get_network_objects(device=10, object_ids=[80681])
        assert isinstance(network_objects, list)
        assert len(network_objects) == 1

        groups = st.get_network_object_groups(15)
        assert len(groups) == 1
        assert groups[0].name == "Panorama-G-1"

        pager = st.get_network_object_rules(12)
        assert isinstance(pager, Pager)
        rules = pager.fetch_all()
        assert len(rules) == 3
        assert rules[0].id == 26
        assert rules[0].name == "HQ access"



from typing import Optional


from pytos2.models import Jsonable
from pytos2.utils import propify, prop


@propify
class Domain(Jsonable):
    id: int = prop(0, converter=int)
    description: Optional[str] = prop("")
    address: Optional[str] = prop("")
    name: str = prop(None)



from datetime import datetime
from enum import Enum

from typing import List, Optional

from ..utils import propify, prop, safe_iso8601_date, kwargify, multiroot_kwargify
from ..models import Jsonable, Link, ObjectReference

from pytos2.secureapp import entrypoint as sa

from pytos2.secureapp.service_object import Service, classify_service_object
from pytos2.secureapp.network_object import NetworkObject, classify_network_object


def classify_history_details(obj):
    _type = {
        "historyConnectionDetailsDTO": HistoryConnectionDetails,
        "historyApplicationDetailsDTO": HistoryApplicationDetails,
        "historyApplicationInterfaceDetailsDTO": HistoryApplicationInterfaceDetails,
        "historyApplicationPackInstanceDetailsDTO": HistoryApplicationPackInstanceDetails,
        "historyNetworkObjectDetailsDTO": HistoryNetworkObjectDetails,
        "historyApplicationInterfaceInstanceDetailsDTO": HistoryApplicationInterfaceInstanceDetails,
        "historyServiceDetailsDTO": HistoryServiceDetails,
    }.get(obj.get("@xsi.type"), HistoryDetails)

    return _type.kwargify(obj)


@propify
class BaseProperties(Jsonable):
    comment: Optional[str] = prop(None)
    name: Optional[str] = prop(None)


@propify
class HistoryDetails(Jsonable):
    class XsiType(Enum):
        HISTORY_CONNECTION_DETAILS_DTO = "historyConnectionDetailsDTO"
        HISTORY_APPLICATION_DETAILS_DTO = "historyApplicationDetailsDTO"
        HISTORY_APPLICATION_INTERFACE_DETAILS_DTO = (
            "historyApplicationInterfaceDetailsDTO"
        )
        HISTORY_APPLICATION_PACK_INSTANCE_DETAILS_DTO = (
            "historyApplicationPackInstanceDetailsDTO"
        )
        HISTORY_NETWORK_OBJECT_DETAILS_DTO = "historyNetworkObjectDetailsDTO"
        HISTORY_APPLICATION_INTERFACE_INSTANCE_DETAILS_DTO = (
            "historyApplicationInterfaceInstanceDetailsDTO"
        )
        HISTORY_SERVICE_DETAILS_DTO = "historyServiceDetailsDTO"

    xsi_type: Optional[XsiType] = prop(None, key="@xsi.type")


@propify
class HistoryApplicationInterfaceDetails(Jsonable):
    new: Optional[BaseProperties] = prop(None)
    previous: Optional[BaseProperties] = prop(None)


@propify
class HistoryApplicationInterfaceInstanceDetails(Jsonable):
    new: Optional[BaseProperties] = prop(None)
    previous: Optional[BaseProperties] = prop(None)


@propify
class HistoryApplicationPackInstanceDetails(Jsonable):
    new: Optional[BaseProperties] = prop(None)
    previous: Optional[BaseProperties] = prop(None)


@propify
class HistoryApplicationDetails(Jsonable):
    ticket: Optional[ObjectReference] = prop(None)
    previous_owner: Optional[ObjectReference] = prop(None)
    new_owner: Optional[ObjectReference] = prop(None)
    removed_editors: List[ObjectReference] = prop(factory=list, flatify="editor")
    removed_viewers: List[ObjectReference] = prop(factory=list, flatify="viewer")


@propify
class ObjectReferenceList(Jsonable):
    source: List[ObjectReference] = prop(factory=list)
    destination: List[ObjectReference] = prop(factory=list)
    service: List[ObjectReference] = prop(factory=list)


def find_in_detail_list(removed, k):
    for r in removed:
        ret = getattr(r, k)
        if ret:
            return ret
    return []


@propify
class HistoryConnectionDetails(HistoryDetails):
    """
    new (BasePropertiesDTO, optional): new base properties,
    previous (BasePropertiesDTO, optional): previous base properties,
    removed ( wrapper,
    source (array[ObjectReferenceDTO], optional): sources removed),
    removed ( wrapper,
    destination (array[ObjectReferenceDTO], optional): destinations removed),
    service (void, optional): services removed
    """

    new: Optional[BaseProperties] = prop(None)
    previous: Optional[BaseProperties] = prop(None)

    added: List[ObjectReferenceList] = prop(factory=list)
    removed: List[ObjectReferenceList] = prop(factory=list)

    @property
    def added_sources(self):
        return find_in_detail_list(self.added, "source")

    @property
    def added_destinations(self):
        return find_in_detail_list(self.added, "destination")

    @property
    def added_services(self):
        return find_in_detail_list(self.added, "service")

    @property
    def removed_sources(self):
        return find_in_detail_list(self.removed, "source")

    @property
    def removed_destinations(self):
        return find_in_detail_list(self.removed, "destination")

    @property
    def removed_services(self):
        return find_in_detail_list(self.removed, "service")


@propify
class HistoryServiceDetails(HistoryConnectionDetails):
    pass


@propify
class HistoryNetworkObjectDetails(HistoryDetails):
    new: Optional[NetworkObject] = prop(None, kwargify=classify_network_object)
    previous: Optional[NetworkObject] = prop(None, kwargify=classify_network_object)
    added: List[ObjectReference] = prop(factory=list)
    removed: List[ObjectReference] = prop(factory=list)


@propify
class HistoryRecord(Jsonable):
    class XsiType(Enum):
        HISTORY_BASE_CONNECTION_DTO = "historyBaseConnectionDTO"
        HISTORY_DEVICE_OBJECT_DTO = "historyDeviceObjectDTO"

    xsi_type: Optional[XsiType] = prop(None, key="@xsi.type")

    date: Optional[datetime] = prop(None, kwargify=safe_iso8601_date)
    object_name: Optional[str] = prop(None)

    # Despite what the API documentation says, this is actually a List of ObjectReference.
    # Make the code we expose to the developer reflect the API documentation.
    user: Optional[ObjectReference] = prop(
        None, kwargify=lambda x: ObjectReference.kwargify(x[0]) if x else None
    )
    type: Optional[str] = prop(None)
    modified_object: Optional[ObjectReference] = prop(None)
    change_description: Optional[str] = prop(None)
    change_details: Optional[HistoryDetails] = prop(
        None, kwargify=classify_history_details
    )

    @classmethod
    def kwargify(cls, obj):
        if obj.get("@xsi.type") == "historyBaseConnectionDTO":
            _obj, kwargs = kwargify(HistoryBaseConnectionRecord, obj)
            return HistoryBaseConnectionRecord(**kwargs)
        elif obj.get("@xsi.type") == "historyDeviceObjectDTO":
            _obj, kwargs = kwargify(HistoryDeviceObjectRecord, obj)
            return HistoryDeviceObjectRecord(**kwargs)
        else:
            _obj, kwargs = kwargify(HistoryRecord, obj)
            return HistoryRecord(**kwargs)
            raise ValueError(
                f"Unknown xsi.type in HistoryRecord: {obj.get('@xsi.type')}"
            )


@propify
class HistoryConnectionBaseSnapshot(Jsonable):
    class XsiType(Enum):
        HISTORY_CONNECTION_DTO = "historyConnectionDTO"

    xsi_type: Optional[XsiType] = prop(None, key="@xsi.type")
    services: List[Service] = prop(
        flatify="service", factory=list, kwargify=classify_service_object
    )
    comment: Optional[str] = prop(None)
    sources: List[NetworkObject] = prop(
        flatify="source", factory=list, kwargify=classify_network_object
    )
    destinations: List[NetworkObject] = prop(
        flatify="destination", factory=list, kwargify=classify_network_object
    )
    name: Optional[str] = prop(None)


@propify
class HistoryBaseConnectionRecord(HistoryRecord):
    xsi_type: Optional[HistoryRecord.XsiType] = prop(
        HistoryRecord.XsiType.HISTORY_BASE_CONNECTION_DTO, key="@xsi.type"
    )

    snapshot: Optional[HistoryConnectionBaseSnapshot] = prop(None)


@propify
class HistoryDeviceObjectRecord(HistoryRecord):
    xsi_type: Optional[HistoryRecord.XsiType] = prop(
        HistoryRecord.XsiType.HISTORY_DEVICE_OBJECT_DTO, key="@xsi.type"
    )

    snapshot: Optional[NetworkObject] = prop(None, kwargify=classify_network_object)
    date: Optional[datetime] = prop(None, kwargify=safe_iso8601_date)
    object_name: Optional[str] = prop(None)



import pytest
import json
import responses

from pytos2.securetrack.time_object import TimeObject
from pytos2.utils import get_api_node


class TestTimeObject:
    @responses.activate
    def test_get_time_objects(self, time_object_mock, st):
        time_objects = st.get_time_objects(2812)

        assert isinstance(time_objects[0], TimeObject)
        assert time_objects[0].id == 388
        assert time_objects[0].name == "Any"
        assert time_objects[0].uid == "{97AEB369-9AEA-11D5-BD16-0090272CCB30}"
        assert time_objects[0].is_global is False
        assert time_objects[0].class_name == "any_object"
        assert len(time_objects[0].time_intervals) == 0

        assert isinstance(time_objects[1], TimeObject)
        assert time_objects[1].id == 391
        assert time_objects[1].name == "Off_Work"
        assert time_objects[1].uid == "{4E4A450C-6A62-44A6-B974-AAC64F8B1DD8}"
        assert time_objects[1].is_global is False
        assert time_objects[1].class_name == "time"
        assert len(time_objects[1].time_intervals) == 2

        assert time_objects[1].time_intervals[0].from_time == "00:00"
        assert time_objects[1].time_intervals[0].to_time == "07:00"
        assert time_objects[1].time_intervals[1].from_time == "20:00"
        assert time_objects[1].time_intervals[1].to_time == "23:59"

        assert isinstance(time_objects[2], TimeObject)
        assert time_objects[2].id == 390
        assert time_objects[2].name == "Weekend"
        assert time_objects[2].uid == "{032FBEA8-A4D4-4134-9035-AA2A56091474}"
        assert time_objects[2].is_global is False
        assert time_objects[2].class_name == "time"
        assert len(time_objects[2].time_intervals) == 0

    @responses.activate
    def test_get_time_objects_by_id(self, time_object_mock, st):
        time_objects = st.get_time_objects(2812, [388])

        assert isinstance(time_objects[0], TimeObject)
        assert time_objects[0].id == 388
        assert time_objects[0].name == "Any"
        assert time_objects[0].uid == "{97AEB369-9AEA-11D5-BD16-0090272CCB30}"
        assert time_objects[0].is_global is False
        assert time_objects[0].class_name == "any_object"
        assert len(time_objects[0].time_intervals) == 0

        time_objects = st.get_time_objects(2812, [388, 390])

        assert isinstance(time_objects[0], TimeObject)
        assert time_objects[0].id == 388
        assert time_objects[0].name == "Any"
        assert time_objects[0].uid == "{97AEB369-9AEA-11D5-BD16-0090272CCB30}"
        assert time_objects[0].is_global is False
        assert time_objects[0].class_name == "any_object"
        assert len(time_objects[0].time_intervals) == 0

        assert isinstance(time_objects[1], TimeObject)
        assert time_objects[1].id == 390
        assert time_objects[1].name == "Weekend"
        assert time_objects[1].uid == "{032FBEA8-A4D4-4134-9035-AA2A56091474}"
        assert time_objects[1].is_global is False
        assert time_objects[1].class_name == "time"
        assert len(time_objects[1].time_intervals) == 0

    @responses.activate
    def test_get_time_objects_by_device(self, time_object_mock, st):
        time_objects = st.get_time_objects_by_device(1)

        assert isinstance(time_objects[0], TimeObject)
        assert time_objects[0].id == 83
        assert time_objects[0].name == "All"
        assert time_objects[0].uid == "{97AEB368-9AEA-11D5-BD16-0090272CCB30}"
        assert time_objects[0].is_global is False
        assert time_objects[0].class_name == "any_object"
        assert len(time_objects[0].time_intervals) == 0

        assert isinstance(time_objects[1], TimeObject)
        assert time_objects[1].id == 87
        assert time_objects[1].name == "Off_Work"
        assert time_objects[1].uid == "{4E4A450C-6A62-44A6-B974-AAC64F8B1DD8}"
        assert time_objects[1].is_global is False
        assert time_objects[1].class_name == "time"
        assert len(time_objects[1].time_intervals) == 2

        assert time_objects[1].time_intervals[0].from_time == "00:00"
        assert time_objects[1].time_intervals[0].to_time == "07:00"
        assert time_objects[1].time_intervals[1].from_time == "20:00"
        assert time_objects[1].time_intervals[1].to_time == "23:59"

    @responses.activate
    def test_device_get_time_objects(self, time_object_mock, devices_mock, st):
        device = st.get_device(1)
        time_objects = device.get_time_objects()

        assert isinstance(time_objects[0], TimeObject)
        assert time_objects[0].id == 83
        assert time_objects[0].name == "All"
        assert time_objects[0].uid == "{97AEB368-9AEA-11D5-BD16-0090272CCB30}"
        assert time_objects[0].is_global is False
        assert time_objects[0].class_name == "any_object"
        assert len(time_objects[0].time_intervals) == 0

        assert isinstance(time_objects[1], TimeObject)
        assert time_objects[1].id == 87
        assert time_objects[1].name == "Off_Work"
        assert time_objects[1].uid == "{4E4A450C-6A62-44A6-B974-AAC64F8B1DD8}"
        assert time_objects[1].is_global is False
        assert time_objects[1].class_name == "time"
        assert len(time_objects[1].time_intervals) == 2

        assert time_objects[1].time_intervals[0].from_time == "00:00"
        assert time_objects[1].time_intervals[0].to_time == "07:00"
        assert time_objects[1].time_intervals[1].from_time == "20:00"
        assert time_objects[1].time_intervals[1].to_time == "23:59"



import pytest
import json
import responses

from pytos2.securetrack.generic_interface import GenericInterface
from pytos2.utils import get_api_node


class TestGenericInterface:
    interface1 = json.load(open("tests/securetrack/json/generic_interfaces/int-1.json"))
    int1 = GenericInterface.kwargify(interface1["GenericInterface"])

    mgmt = json.load(open("tests/securetrack/json/generic_interfaces/mgmt-1.json"))
    ipv4Interfaces = [
        GenericInterface.kwargify(intfc) for intfc in mgmt["GenericInterfaces"]
    ]
    mgmt5 = json.load(open("tests/securetrack/json/generic_interfaces/mgmt-5.json"))
    ipv6Interfaces = [
        GenericInterface.kwargify(intfc) for intfc in mgmt5["GenericInterfaces"]
    ]

    @responses.activate
    def test_add_generic_interfaces_201(self, st):
        """Add one or multiple interfaces, we only support adding one though"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/interface",
            status=201,
        )

        """Add single interface"""
        newInt1 = st.add_generic_interface(
            device=self.int1["device_id"],
            name=self.int1["name"],
            ip=self.int1["ip"],
            mask=self.int1["mask"],
            vrf=self.int1["vrf"],
            mpls=self.int1["mpls"],
            unnumbered=self.int1["unnumbered"],
            type=self.int1["type"],
        )
        assert newInt1 is None

    @responses.activate
    def test_add_generic_interfaces_400(self, st):
        """POST bad request"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/interface",
            status=400,
        )

        with pytest.raises(ValueError) as exception:
            st.add_generic_interface(
                device=self.int1["device_id"],
                name=self.int1["name"],
                ip=self.int1["ip"],
                mask=self.int1["mask"],
                vrf=self.int1["vrf"],
                mpls=self.int1["mpls"],
                unnumbered=self.int1["unnumbered"],
                type=self.int1["type"],
            )
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_add_generic_interfaces_404(self, st):
        """POST mgmt id not found"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/interface",
            status=404,
            json={"result": {"message": "Management Not Found"}},
        )

        with pytest.raises(ValueError) as exception:
            st.add_generic_interface(
                device=self.int1["device_id"],
                name=self.int1["name"],
                ip=self.int1["ip"],
                mask=self.int1["mask"],
                vrf=self.int1["vrf"],
                mpls=self.int1["mpls"],
                unnumbered=self.int1["unnumbered"],
                type=self.int1["type"],
            )
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_interface(self, st, generic_interfaces_mock):
        """Get interface by id"""
        mockRes = get_api_node(self.interface1, "GenericInterface")
        interface = GenericInterface.kwargify(mockRes)
        intByInt = st.get_generic_interface(1)
        assert intByInt == interface

        with pytest.raises(ValueError) as exception:
            st.get_generic_interface(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.get_generic_interface("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_interfaces(self, st, generic_interfaces_mock):
        """Get all interfaces associated with management id"""
        interfaces = [
            GenericInterface.kwargify(d)
            for d in get_api_node(self.mgmt, "GenericInterfaces", listify=True)
        ]
        intsByInt = st.get_generic_interfaces(1)
        assert intsByInt == interfaces

        with pytest.raises(ValueError) as exception:
            st.get_generic_interfaces(404)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_update_generic_interfaces_200(self, st):
        """Update one or multiple interfaces"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/interface",
            status=200,
        )

        """Update single interface"""
        newInterface = st.update_generic_interface(self.int1)
        assert newInterface is None

    @responses.activate
    def test_update_generic_interfaces_400(self, st):
        """PUT bad request"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/interface",
            status=400,
        )

        with pytest.raises(ValueError) as exception:
            st.update_generic_interface(self.int1)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_update_generic_interfaces_404(self, st):
        """PUT mgmt id not found"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/interface",
            status=404,
        )

        with pytest.raises(ValueError) as exception:
            st.update_generic_interface(self.int1)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_interface(self, st, generic_interfaces_mock):
        """Delete interface by interface id"""
        intByInt = st.delete_generic_interface(1)
        assert intByInt is None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_interface(404)
        assert "Not Found" in str(exception.value)



import pytest
import responses

from requests.exceptions import HTTPError
from . import conftest  # noqa
from pytos2.securechange.entrypoint import Scw


class TestClient(object):
    def test_with_args(self, scw_api):
        assert scw_api.hostname == "198.18.0.1"
        assert scw_api.username == "username"
        assert scw_api.password == "password"

    def test_with_missing_args(self):
        with pytest.raises(ValueError) as e:
            Scw()
        assert "hostname argument must be provided" in str(e.value)


class TestClientMethods(object):
    @responses.activate
    def test_get_ticket_not_found(self, scw):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/100000000",
            match_querystring=False,
            body=open("tests/securechange/json/ticket/ticket_not_found.json").read(),
            content_type="application/xml",
            status=404,
        )
        with pytest.raises(HTTPError):
            scw.get_ticket(100000000)

    @responses.activate
    def test_get_ticket_error(self, scw):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/100000000",
            match_querystring=False,
            content_type="application/xml",
            status=500,
        )
        # XXX TODO: implement
        pass


class TestAPIMethods:
    @responses.activate
    def test_get_workflow(self, scw_api, all_fields_workflow_json):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/workflows",
            match_querystring=False,
            json=all_fields_workflow_json,
        )

        assert (
            scw_api.get_workflow(336).json()
            == scw_api.get_workflow(name="AR all fields").json()
        )

        with pytest.raises(TypeError):
            scw_api.get_workflow(336, name="AR all fields")

    @responses.activate
    def test_reassign_task(self, scw_api):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1/steps/1234/tasks/1234/reassign/1",
        )
        r = scw_api.reassign_task(1, 1234, 1234, 1)
        assert r.ok

    @responses.activate
    def test_put_task(self, scw_api):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1/steps/current/tasks/1234",
        )
        r = scw_api.put_task(1, 1234, {"id": 1234})
        assert r.ok

    @responses.activate
    def test_put_fields(self, scw_api, all_fields_field_json):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1/steps/current/tasks/1234/fields",
        )
        r = scw_api.put_fields(1, 1234, all_fields_field_json)
        assert r.ok

    @responses.activate
    def test_cancel_ticket(self, scw_api):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1/cancel",
        )
        r = scw_api.cancel_ticket(1, 123)
        assert r.ok

    @responses.activate
    def test_confirm_ticket(self, scw_api):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1/confirm",
        )
        r = scw_api.confirm_ticket(1)
        assert r.ok



from enum import Enum
import json

import pytest  # type: ignore
from netaddr import IPNetwork, IPAddress, IPRange  # type: ignore

from pytos2.utils import (
    snakecase_to_upper_camelcase,
    NoInstance,
    singleton,
    setup_logger,
    propify,
    prop,
    kwargify,
    get_api_node,
    deprecated_property,
)
from typing import Dict, List, Optional, Union
from pytos2.models import Jsonable, UnMapped


@propify
class DeprecatedPropertyClass(Jsonable):
    test_property: str = prop("TEST VALUE")

    deprecated_property = deprecated_property("deprecated_property", "test_property")


@propify
class Child(Jsonable):
    name: str = prop()
    nested: List["Child"] = prop(factory=list)


@propify
class DictKwargify(Jsonable):
    name: str = prop()
    properties_dict: Dict = prop(factory=dict)


@propify
class Parent(Jsonable):
    class Meta(Enum):
        ROOT = "root"

    class Things(Enum):
        THIS = "this"
        THAT = "that"

    class Maybe(Enum):
        MAYBE = "MAYBE"

    class Flub(Enum):
        FLARG = "flarg"

    class Prop(Enum):
        FOOBAR = "$#@%$"

    foobar: Optional[str] = prop(key=Prop.FOOBAR.value)
    keys: List[str] = prop(flatify="key")
    empties: List[str] = prop(flatify="empty")
    maybe: Optional[Maybe] = prop()
    thing: Things = prop()
    flub: Flub = prop()
    no_json: Optional[bool] = prop(None, jsonify=False)
    data: Optional[dict] = None
    missing_list: List[str] = prop(factory=list)
    missing: Optional[str] = prop(None)
    wrong: Union[str, int] = prop()
    netty: IPNetwork = prop(kwargify=False)
    addy: IPAddress = prop()
    rangy: IPRange = prop(kwargify=False)
    children: List[Child] = prop(factory=list)

    @classmethod
    def kwargify(cls, obj):
        _obj, kwargs = kwargify(cls, obj)
        kwargs["netty"] = IPNetwork(f"{_obj['network']}/{_obj['mask']}")
        kwargs["rangy"] = IPRange(*_obj["rangy"].split("-"))
        return cls(**kwargs)


@pytest.fixture
def parent():
    return Parent.kwargify(
        {
            "root": {
                "keys": {"key": "value"},
                "empties": "",
                "$#@%$": "eff word",
                "ignore": "sad, many such cases",
                "no_json": "hare today",
                "flub": "Negatrov",
                "thing": "this",
                "extra": "MIA",
                "maybe": "yes",
                "wrong": 12,
                "network": "1.2.3.0",
                "mask": "255.255.255.0",
                "addy": "1.2.3.4",
                "rangy": "1.1.1.1-2.2.2.2",
                "children": [
                    {
                        "name": "child1",
                        "nested": [{"name": "child2"}, {"name": "child3"}],
                    }
                ],
            }
        }
    )


def test_kwargify(parent):
    assert isinstance(parent.keys, list)
    assert isinstance(parent.empties, list)
    assert parent.foobar == "eff word"
    with pytest.raises(AttributeError):
        parent.ignore
    assert parent.thing is parent.Things.THIS
    assert parent.flub == "Negatrov"
    assert parent["flub"] == "Negatrov"


def test_kwargify_dict(parent):
    # This ensures that the typing.Dict is properly handled.
    dk = DictKwargify.kwargify({"name": "test", "properties_dict": {}})
    assert dk.name == "test"
    assert dk.properties_dict == {}


def test_kwargify_missing_props(parent):
    bad = parent.data
    del bad["root"]["thing"]
    with pytest.raises(ValueError):
        Parent.kwargify(bad)


def test_json(parent):
    assert parent._json == {
        "keys": {"key": ["value"]},
        "empties": {"empty": []},
        "$#@%$": "eff word",
        "thing": "this",
        "maybe": "yes",
        "flub": "Negatrov",
        "missing_list": [],
        "wrong": 12,
        "netty": IPNetwork("1.2.3.0/24"),
        "addy": IPAddress("1.2.3.4"),
        "rangy": IPRange("1.1.1.1", "2.2.2.2"),
        "children": [
            {
                "name": "child1",
                "nested": [
                    {"name": "child2", "nested": []},
                    {"name": "child3", "nested": []},
                ],
            }
        ],
    }


def test_json_override(parent):
    override = {"some": "dic"}
    parent._json = override
    assert parent._json is override


def test_deprecated_property():
    p_class = DeprecatedPropertyClass(test_property="TEST")
    assert p_class.deprecated_property == "TEST"

    p_class.deprecated_property = "ANOTHER VALUE"
    assert p_class.test_property == "ANOTHER VALUE"


@singleton
class Single(object):
    pass


def test_unmapped():
    _json = {"some": "thing"}
    obj = UnMapped.kwargify(_json)
    assert obj._json == _json


class TestHelpers:
    def test_singleton(self):
        single1 = Single()
        single2 = Single()
        assert single1 is single2

    def test_no_instance_obj(self):
        name, msg = "test", "Test error msg"
        no_instance = NoInstance(name, msg)
        with pytest.raises(RuntimeError):
            no_instance.nothing

    def test_snakecase_to_upper_camelcase(self):
        assert (
            snakecase_to_upper_camelcase("snek_case_is_awesome") == "SnekCaseIsAwesome"
        )


class TestGetApiNode:
    empty: dict = {}
    empty_nested: dict = {"things": {}}
    blank_string_nested: dict = {"things": ""}
    flat_nested: dict = {"things": {"thing": {"key": "val"}}}
    list_nested: dict = {"things": {"thing": [{"key": "val"}]}}

    @pytest.mark.parametrize(
        "obj, path, val",
        (
            (empty, "things.thing", None),
            (empty_nested, "things.thing", None),
            (blank_string_nested, "things.thing", None),
            (flat_nested, "things.thing.key", "val"),
        ),
    )
    def test_empty(self, obj, path, val):
        assert get_api_node(obj, path) == val

    @pytest.mark.parametrize(
        "obj, path, val",
        (
            (empty, "things.thing", []),
            (empty_nested, "things.thing", []),
            (blank_string_nested, "things.thing", []),
            (flat_nested, "things.thing", [{"key": "val"}]),
        ),
    )
    def test_listify(self, obj, path, val):
        assert get_api_node(obj, path, listify=True) == val

    @pytest.mark.parametrize(
        "obj, path, val",
        (
            (empty, "things.thing", "default"),
            (empty_nested, "things.thing", "default"),
            (blank_string_nested, "things.thing", "default"),
            (flat_nested, "things.thing", [{"key": "val"}]),
        ),
    )
    def test_default(self, obj, path, val):
        assert get_api_node(obj, path, listify=True, default="default") == val

    def test_list_key_exception(self):
        with pytest.raises(ValueError) as excinfo:
            get_api_node(self.list_nested, "things.thing.key")
        assert "obj['things']['thing']" in str(excinfo.value)


# class TestLogging:
#     def test_with_file_arg(self, tmp_path):
#         lo = setup_logger("test", tmp_path)
#         lo.error("TEST MESSAGE")
#         msg = json.loads(open(tmp_path / "test.log.json").readline())
#         assert msg["message"] == "TEST MESSAGE"
#
#     def test_duplicate_logger(self):
#         lo = setup_logger("test")
#         assert lo == setup_logger("test")
#
#     def test_stdout_logger(self, caplog):
#         lo = setup_logger("out")
#         lo.error("TEST MESSAGE", extra={"ignore": "ignore"})
#         assert "TEST MESSAGE" in caplog.records[0].message
#
#     def test_dict_formatter(self, caplog):
#         lo = setup_logger("out")
#         lo.error({"message": "TEST MESSAGE", "hide": "HIDE"})
#         assert "TEST MESSAGE" in caplog.records[0].message



import pytest
import json
import responses
from . import conftest

from pytos2.securetrack.security_policy import (
    SecurityPolicy,
    SecurityPolicyDeviceMapping,
    SecurityPolicyInterface,
    SecurityPolicyMatrixType,
    SecurityZoneMatrix,
    CSVData,
    InterfacesManualMappings,
    InterfaceUserMapping,
    ZoneUserAction,
)


class TestSecurityPolicy:
    @responses.activate
    def test_get_security_policies(self, security_policy_mock, st):
        security_policies = st.get_usp_policies()
        first = security_policies[0]
        assert first.id == 7131354762492230143
        assert first.name == "27001"
        assert first.domain_id == "1"
        assert first.domain_name == "Default"
        assert first.description is None
        assert first.type == SecurityPolicyMatrixType.SECURITY_ZONE_MATRIX
        assert (
            security_policies[13].description
            == "Aligning trust and networks to the Trusted Internet Connections Model"
        )

        security_policies = st.get_usp_policies(get_global=True)
        first = security_policies[0]
        assert first.id == 7131354762492230143
        assert first.name == "27001"
        assert first.domain_id == "1"
        assert first.domain_name == "Default"
        assert first.type == SecurityPolicyMatrixType.SECURITY_ZONE_MATRIX

    @responses.activate
    def test_get_security_policies_ignoreSecureTrack2Data(
        self, security_policy_mock, st
    ):
        security_policies = st.get_usp_policies(aurora_data=False)
        first = security_policies[0]
        assert first.id == 1
        assert first.name == "Corporate Matrix (Physical + AWS)"
        assert first.domain_id == "1"
        assert first.domain_name == "Default"
        assert first.description == ""

    @responses.activate
    def test_export_security_policy(self, security_policy_mock, st):
        csv_string = st.export_usp_policy(7131354762492230143)
        assert isinstance(csv_string, str)

        with pytest.raises(ValueError) as exception:
            st.export_usp_policy(404)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_usp_map(self, security_policy_mock, st):
        security_policy_device_mapping = st.get_usp_map(5)
        assert security_policy_device_mapping.device_id == 5
        assert isinstance(security_policy_device_mapping.affiliated_interfaces, list)
        assert isinstance(security_policy_device_mapping.interfaces, list)
        interf = security_policy_device_mapping.interfaces[1]
        assert interf.name == "ge-0/0/1.1"
        assert interf.zones[0] == "Virtual_DC-04"

    @responses.activate
    def test_delete_security_policy(self, security_policy_mock, st):
        deleted = st.delete_usp_policy(7131354762492230143)
        assert deleted is None

        with pytest.raises(ValueError) as exception:
            st.delete_usp_policy(404)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_add_usp_map(self, security_policy_mock, st, zones_mock):
        resp = st.add_usp_map(device=5, interface_name="ge-0/0/0.0", zone=147)
        assert resp is None

        with pytest.raises(ValueError) as exception:
            st.add_usp_map(404, "", -1)
        assert "Not Found" in str(exception.value)

        resp = st.add_usp_map(
            device=5, interface_name="ge-0/0/0.0", zone="SomeTestZone"
        )
        assert resp is None

        with pytest.raises(ValueError) as e:
            st.add_usp_map(
                device=5, interface_name="ge-0/0/0.0", zone="SomeNonexistentZone"
            )
        assert "No matching zones" in str(e.value)

    @responses.activate
    def test_delete_usp_map(self, security_policy_mock, st):
        resp = st.delete_usp_map(device=5, interface_name="ge-0/0/0.0", zone=147)
        assert resp is None

        with pytest.raises(ValueError) as exception:
            st.delete_usp_map(404, "", -1)
        assert "Not Found" in str(exception.value)



from os import getenv
from typing import Iterable, Iterator, Optional, TypeVar, Any
from urllib.parse import urljoin, urlparse
import logging

from oauthlib.oauth2 import TokenExpiredError
import requests
from requests import JSONDecodeError, RequestException, Response
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from traversify import Traverser  # type: ignore

from pytos2.utils import setup_logger, get_api_node
from .models import OAuthToken
from requests_oauthlib import OAuth2Session

import warnings

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
LOGGER = setup_logger("api")
REQUESTS_LOGGER = logging.getLogger("requests.packages.urllib3")


class SdkError(ValueError):
    def __init__(self, message):
        super().__init__(message)


class ApiError(SdkError):
    def __init__(self, message):
        super().__init__(message)


class GraphqlError(ApiError):
    def __init__(self, errors):
        self.errors = errors
        super().__init__(f"GraphQL Error: {errors}")


class BaseAPI:
    def __init__(self, session):
        self.session = session

    def handle_response(self, r, fn_name, action=None) -> Response:
        """
        Use this method to handle responses from the API

        :param r: response Object
        :param fn_name: name of the function that made the request
        :param expects_json: if the response is expected to be json
        :return: response or json
        """
        if not action:
            action = fn_name.split("_")[0]

        if not r.ok:
            try:
                msg = r.json()
                message = msg.get("result", {}).get("message", "")
                code = msg.get("result", {}).get("code", "")
            except JSONDecodeError as e:
                raise ApiError(
                    f"{fn_name}: Failed to {action} resource with status {r.status_code}: {r.text}"
                ) from e
            raise ApiError(
                f"{fn_name}: Failed to {action} resource with status {r.status_code}: ({code}) {message}"
            )
        return r

    def handle_json(self, r, fn_name, action="get") -> dict:
        r = self.handle_response(r, fn_name, action=action)
        try:
            return r.json()
        except JSONDecodeError as e:
            raise ApiError(f"{fn_name}: Failed to decode response: {r.text}") from e

    def handle_creation(self, r, fn_name, *, cls, action="add", warnme=True) -> Any:
        r = self.handle_response(r, fn_name)
        if r.status_code == 201:
            loc = r.headers["Location"]

            res = self.session.get(loc)
            _json = self.handle_json(res, fn_name, action=action)

            if warnme and (not hasattr(cls, "Meta") or not hasattr(cls.Meta, "ROOT")):
                warnings.warn(
                    f"{cls.__name__} does not have a Meta.ROOT attribute. This may cause issues in handle_creation. If you are seeing this warning, please add a Meta class with a ROOT attribute to {cls.__name__}. If you know what you're doing, you may pass warnme=True to this function."
                )

            if (
                hasattr(cls, "Meta")
                and hasattr(cls.Meta, "ROOT")
                and cls.Meta.ROOT.value in _json
            ):
                _json = _json[cls.Meta.ROOT.value]

            if isinstance(_json, list):
                kwargified = [cls.kwargify(i) for i in _json]
                if len(kwargified) == 1:
                    return kwargified[0]
                return kwargified

            return cls.kwargify(_json)
        else:
            resource_descriptor = getattr(cls.Meta, "ENTITY", "resource")
            err = f"{fn_name}: Failed to add {resource_descriptor} with status {r.status_code}: {r.text}"
            raise ApiError(err)


class APISession(requests.Session):
    """
    Inits a prefixed API session

    :param str base_url: The url for the tos host
    :param str username: username
    :param str password: password
    :param str api_path: api path to prefix all requests
    :param bool verify: sets ssl strict verification for all requests
    """

    def __init__(
        self,
        hostname,
        scheme="https",
        username=None,
        password=None,
        api_path="",
        verify=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if username is not None and password is not None:
            self.username = username
            self.password = password

        if kwargs.get("headers") is None:
            self.headers.update({"Accept": "application/json, */*"})

        self.verify = verify
        host_base = urlparse(hostname)

        self.base_url = urljoin(
            f"{scheme}://" + (host_base.netloc or host_base.path),
            api_path.strip("/") + "/",
        )

        self.scheme = scheme
        self.hostname = host_base.netloc or host_base.path

    @property
    def auth(self):
        return self.username, self.password

    @auth.setter
    def auth(self, value):
        self.username, self.password = value if value else (None, None)

    def request(self, method, url, *args, **kwargs):
        """Overrides all requests to prefix url, not intended to be used directly"""

        # Make sure we have a sensible timeout set
        if not kwargs.get("timeout", None):
            kwargs["timeout"] = 300
        REQUESTS_LOGGER.debug(f"Request kwargs {kwargs}")

        try:
            response = super().request(
                method, urljoin(self.base_url, url), *args, **kwargs
            )
        except RequestException as e:
            raise ApiError(f"Failed to make request: {e}") from e
        REQUESTS_LOGGER.debug(f"Response body {response.text}")

        return response


class OAuth2APISession(OAuth2Session):
    """
    Inits a prefixed API session with oauth2 support


    :param str base_url: The url for the tos host
    :param str username: username
    :param str password: password
    :param str api_path: api path to prefix all requests
    :param bool verify: sets ssl strict verification for all requests
    """

    def __init__(
        self,
        hostname,
        scheme="https",
        username=None,
        password=None,
        client_id=None,
        api_path="",
        verify=False,
        **kwargs,
    ):
        super().__init__(client_id, **kwargs)
        if username is not None and password is not None:
            self.username = username
            self.password = password

        if kwargs.get("headers") is None:
            self.headers.update({"Accept": "application/json, */*"})

        self.verify = verify
        host_base = urlparse(hostname)

        self.base_url = urljoin(
            f"{scheme}://" + (host_base.netloc or host_base.path),
            api_path.strip("/") + "/",
        )

        self.scheme = scheme
        self.hostname = host_base.netloc or host_base.path
        self.client_id = client_id

    @property
    def token_url(self) -> str:
        return f"{self.scheme}://{self.hostname}/auth/realms/tufin-realm/protocol/openid-connect/token"

    @property
    def authenticated(self) -> bool:
        return bool(self.access_token)

    def _fetch_token(self) -> OAuthToken:
        data = {
            "grant_type": "password",
            "username": self.username,
            "password": self.password,
            "client_id": self.client_id,
        }

        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        res = requests.post(self.token_url, data=data, headers=headers, verify=False)
        res.raise_for_status()

        self.token = res.json()

        return self.token

    def request(self, method: str, url: str, *args, **kwargs):
        if not self.authenticated:
            self.token = self._fetch_token()

        if self.authenticated:
            kwargs["headers"] = {
                "Authorization": f"Bearer {self.token['access_token']}"
            }

        if not url.startswith("http"):
            url = urljoin(self.base_url, url)

        try:
            response = super().request(
                method, urljoin(self.base_url, url), *args, **kwargs
            )
        except TokenExpiredError:
            self.token = self._fetch_token()
            response = super().request(
                method, urljoin(self.base_url, url), *args, **kwargs
            )

        return response


def get_app_api_session(
    app, hostname=None, username=None, password=None, session_cls=APISession, **kwargs
):
    app_identifier = app.Meta.APP.value.upper()
    app_path = app.Meta.PATH.value
    username_key = "{}_API_USERNAME".format(app_identifier)
    password_key = "{}_API_PASSWORD".format(app_identifier)
    hostname_key = "{}_HOSTNAME".format(app_identifier)

    tos2_env_host_key = "{}_HOST".format(app.Meta.TOS2_ENV.value)
    tos2_env_port_key = "{}_PORT".format(app.Meta.TOS2_ENV.value)

    tos2_host = getenv(tos2_env_host_key, None)
    tos2_port = getenv(tos2_env_port_key, None)

    is_tos2 = getenv(tos2_env_host_key, False)
    scheme = "https"

    hostname = hostname or tos2_host or getenv(hostname_key) or getenv("TOS_HOSTNAME")
    if not hostname:
        raise ValueError(
            f"hostname argument must be provided if {hostname_key} or TOS_HOSTNAME environment variable is not set"
        )
    username = username or getenv(username_key)
    if not username:
        raise ValueError(
            f"username argument must be provided if {username_key} environment variable is not set"
        )
    password = password or getenv(password_key)
    if not password:
        raise ValueError(
            f"password argument must be provided if {password_key} environment variable is not set"
        )

    if is_tos2:
        try:
            tos2_port = int(tos2_port)
        except ValueError:
            tos2_port = 443
        scheme = "https" if tos2_port == 443 else "http"

    return (
        hostname,
        username,
        password,
        session_cls(
            hostname=hostname,
            scheme=scheme,
            username=username,
            password=password,
            api_path=app_path,
            **kwargs,
        ),
    )


def resultify_response(response):
    if response.ok:
        return response
    else:
        response.raise_for_status()


def traversify_response(response):
    if response.ok:
        return Traverser(response)
    else:
        response.raise_for_status()


def boolify(params: dict) -> dict:
    return {k: str(v).lower() if isinstance(v, bool) else v for k, v in params.items()}


T = TypeVar("T")


class Pager(Iterable[T]):
    """
    This class abstracts away Tufin's standard pagination. If you are using the SDK,
    this can be consumed like a list.

    User Example:
    ```
    pager = st.get_applications()
    for app in pager:
        pass

    # Or
    pager = st.get_applications()
    apps = pager.fetch_all()
    # apps is now a list of all applications

    # Or
    pager = st.get_applications()
    app = pager[0]
    some_slice = pager[1:10]
    ```

    """

    def __init__(
        self,
        api,
        url,
        api_node,
        fn_name,
        kwargify,
        page_size=2000,
        start=None,
        stop=None,
        step=None,
        params: Optional[dict] = None,
        is_slice=False,
    ):
        """
        Create a new Pager object.

        :param api: The API object to use for requests
        :param url: The URL to request: i.e. "applications"
        :param api_node: The root node of the response JSON, i.e. "applications.application"
        :param fn_name: The name of the function that made the request -- This is used for error messages
        :param kwargify: A function to convert an element of the response into a Jsonable object.
        :param page_size: The number of items to request per page, default 2000
        :param start: The index to start at, default 0
        :param stop: The index to stop at, default None. You do not need to set this during regular use. It is used internally for slices.
        :param step: Increment the internal index by this amount, default 1
        :param params: Additional parameters to pass to the request
        :param is_slice: Internal flag to determine if this is a slice, default False
        """

        self.api = api
        self.api_node = api_node
        self.fn_name = fn_name
        self.page_size = page_size
        self.url = url
        self.params = params

        self.root_node = api_node.split(".")[0]
        self.kwargify = kwargify

        self.start = start or 0
        self.stop = stop
        self.step = step or 1

        self._items = None
        self._cur_idx = self.start
        self._cur_page_idx = 0
        self._cur_offset = start or 0
        self._next_offset = None
        self._total = None
        self._is_slice = is_slice

        self._accrued = []

        self._prime()

    def __repr__(self):
        item_count = len(self)
        noun = "item" if item_count == 1 else "items"
        length = f"{item_count} {noun}" if self._items is not None else "unknown"
        return f"{self.__class__.__name__}({self.url}, items=[... {length}])"

    def __iter__(self):
        return self

    def _accrue(self):
        if self._items is None or self._cur_page_idx >= len(self._items):
            # Determine whether we should even bother fetching the next page
            if (
                self._total is not None
                and self._cur_offset + self._cur_page_idx >= self._total
            ):
                return False
            elif self._items and len(self._items) < self.page_size:
                return False

            self._cur_offset += self._cur_page_idx
            self._items, total = self._fetch_page(self._cur_offset)
            if total:
                self._total = total

            self._cur_page_idx = 0
            if not self._items:
                return False

        if self.stop and self._cur_page_idx + self._cur_offset >= self.stop:
            return False

        item = self._items[self._cur_page_idx]
        self._cur_page_idx += self.step

        if self.kwargify:
            item = self.kwargify(item)

        self._accrued.append(item)
        return True

    def __next__(self) -> T:
        while self._cur_idx - self.start >= len(self._accrued):
            if not self._accrue():
                break

        if (self._cur_idx - self.start) < len(self._accrued):
            item = self._accrued[self._cur_idx - self.start]
            self._cur_idx += 1
            return item
        else:
            raise StopIteration

    def _prime(self):
        self._items, total = self._fetch_page()
        if total:
            self._total = total

    def __len__(self):
        if self._total is None:
            self._prime()

        if self._total:
            if self.stop and self._total > self.stop:
                return int((self.stop - self.start) / self.step)

            return int((self._total - self.start) / self.step)
        if not self._items:
            return 0

        size = max(len(self._items), len(self._accrued))
        return int(size / self.step)

    def __eq__(self, other):
        if isinstance(other, list):
            return self.fetch_all() == other
        elif isinstance(other, Pager):
            return self.fetch_all() == other.fetch_all()
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def _new_pager(self, start, stop, step, is_slice):
        return Pager(
            self.api,
            self.url,
            self.api_node,
            self.fn_name,
            self.kwargify,
            page_size=self.page_size,
            start=start,
            stop=stop,
            step=step,
            params=self.params,
            is_slice=is_slice,
        )

    def __getitem__(self, idx):
        if self._is_slice and isinstance(idx, slice):
            while self._accrue():
                pass

            return self._accrued[idx]

        if isinstance(idx, slice):
            if idx.step and idx.step < 0:
                raise IndexError("Negative steps are not supported")

            if idx.start and idx.start < 0:
                raise IndexError("Negative start indices are not supported")

            if idx.stop and idx.stop < 0:
                raise IndexError("Negative stop indices are not supported")

            return self._new_pager(
                start=self.start + (idx.start or 0),
                stop=idx.stop + self.start if idx.stop else None,
                step=idx.step,
                is_slice=True,
            )

        if idx < 0:
            while self._accrue():
                pass

            return self._accrued[idx]
        if self._total is None:
            self._prime()

        if self._total and idx >= self._total:
            raise IndexError("Index out of range")

        while self._accrue() and idx >= len(self._accrued):
            pass

        if idx >= len(self._accrued):
            raise IndexError("Index out of range")

        return self._accrued[idx]

    def _params(self, start=None):
        params = (self.params or {}).copy()
        params["start"] = start or self.start or 0

        if self.stop and self.stop - self.start < self.page_size:
            params["count"] = (self.stop - self.start) + 1
        else:
            params["count"] = self.page_size

        params["count"] = self.page_size
        if params["start"] == self.start:
            params["get_total"] = "true"
        return params

    def _fetch_page(self, start=None):
        params = self._params(start)

        response = self.api.session.get(self.url, params=params)
        response_json = self.api.handle_json(response, self.fn_name)
        if params["start"] == self.start:
            total = response_json[self.root_node].get("total", None)
        else:
            total = None

        items = get_api_node(response_json, self.api_node, listify=True)
        return items, total

    def fetch_all(self):
        """
        Fetch all items represented by the pager.
        """

        while self._accrue():
            pass

        return self._accrued



import pytest
from traversify import Traverser
import responses


class TestRelatedRules:
    @responses.activate
    def test_related_rules_results(self, related_rules_mock, scw):
        ticket = scw.get_ticket(49)
        results = ticket.last_task.access_request.related_rules_result.get_results()

        access_request1 = results.access_requests[0]
        assert len(results.access_requests) == 1
        assert access_request1.ar == 1
        assert len(access_request1.devices) == 1
        assert access_request1.devices[0].management_id == 15

        assert len(access_request1.devices[0].bindings) == 1
        assert len(access_request1.devices[0].bindings[0].related_rules) == 2

        related_rule1 = access_request1.devices[0].bindings[0].related_rules[0]
        assert related_rule1.ignore
        assert related_rule1.rule.uid == "{cc7908e5-8780-4357-a0d2-4fa2fd9180be}"
        assert not related_rule1.rule.is_disabled
        assert related_rule1.rule.rule_number == 140
        assert related_rule1.rule.from_zone == "port3"
        assert related_rule1.rule.to_zone == "port2"
        assert related_rule1.rule.track["track_level"] == "LOG"
        assert related_rule1.rule.install_ons == []
        assert related_rule1.rule.communities == []
        assert related_rule1.rule.times == []

        assert len(related_rule1.rule.sources) == 1
        assert (
            related_rule1.rule.sources[0].uid
            == "{e08566b5-184e-463d-b6f5-de44aeede838}"
        )
        assert related_rule1.rule.sources[0].name == "Host_10.103.0.42"
        assert related_rule1.rule.sources[0].display_name == "Host_10.103.0.42"
        assert related_rule1.rule.sources[0].class_name.value == "host_plain"
        assert related_rule1.rule.sources[0].origin == "DEVICE"
        assert not related_rule1.rule.sources[0].is_global
        assert not related_rule1.rule.sources[0].is_implicit
        assert not related_rule1.rule.sources[0].is_shared
        assert related_rule1.rule.sources[0].comment == ""
        assert related_rule1.rule.sources[0].id == "66c7421a9591a744946348da"
        assert related_rule1.rule.sources[0].version_id == 53
        assert related_rule1.rule.sources[0].device_type.value == "Fortinet"
        assert related_rule1.rule.sources[0].ip_type.value == "IPV4"
        assert related_rule1.rule.sources[0].referenced == "UNKNOWN"
        assert not related_rule1.rule.sources[0].installable_target

        assert len(related_rule1.rule.destinations) == 1
        assert (
            related_rule1.rule.destinations[0].uid
            == "{2c8c67e6-d1c2-48ee-a821-3e79b736d2d7}"
        )
        assert related_rule1.rule.destinations[0].name == "Host_10.234.10.42"
        assert related_rule1.rule.destinations[0].display_name == "Host_10.234.10.42"
        assert related_rule1.rule.destinations[0].class_name.value == "host_plain"
        assert related_rule1.rule.destinations[0].origin == "DEVICE"
        assert not related_rule1.rule.destinations[0].is_global
        assert not related_rule1.rule.destinations[0].is_implicit
        assert not related_rule1.rule.destinations[0].is_shared
        assert related_rule1.rule.destinations[0].comment == ""
        assert related_rule1.rule.destinations[0].id == "66c7421a9591a744946348d6"
        assert related_rule1.rule.destinations[0].version_id == 53
        assert related_rule1.rule.destinations[0].device_type.value == "Fortinet"
        assert related_rule1.rule.destinations[0].ip_type.value == "IPV4"
        assert related_rule1.rule.destinations[0].referenced == "UNKNOWN"
        assert not related_rule1.rule.destinations[0].installable_target

        assert len(related_rule1.rule.destination_services) == 1
        assert (
            related_rule1.rule.destination_services[0].uid
            == "{c75809f4-7e46-4f4b-954d-b46cb86d64fb}"
        )
        assert related_rule1.rule.destination_services[0].name == "TCP_441"
        assert related_rule1.rule.destination_services[0].display_name == "TCP_441"
        assert related_rule1.rule.destination_services[0].origin == "DEVICE"
        assert not related_rule1.rule.destination_services[0].is_global
        assert not related_rule1.rule.destination_services[0].implicit
        assert not related_rule1.rule.destination_services[0].shared
        assert related_rule1.rule.destination_services[0].timeout == 0
        assert related_rule1.rule.destination_services[0].comment == ""
        assert related_rule1.rule.destination_services[0].version_id == 53
        assert related_rule1.rule.destination_services[0].referenced.value == "UNKNOWN"
        assert related_rule1.rule.destination_services[0].type_on_device == ""
        assert not related_rule1.rule.destination_services[0].negate
        assert not related_rule1.rule.destination_services[0].match_for_any

        assert (
            related_rule1.intersecting_objects.found_in_source.uid
            == "{e08566b5-184e-463d-b6f5-de44aeede838}"
        )
        assert (
            related_rule1.intersecting_objects.found_in_destination.uid
            == "{2c8c67e6-d1c2-48ee-a821-3e79b736d2d7}"
        )
        assert (
            related_rule1.intersecting_objects.found_in_service.uid
            == "{c75809f4-7e46-4f4b-954d-b46cb86d64fb}"
        )

        related_rule2 = access_request1.devices[0].bindings[0].related_rules[1]
        assert related_rule2.ignore
        assert related_rule2.rule.uid == "{5fc5ca91-805e-4fe3-baa0-2441f27b1d9a}"
        assert not related_rule2.rule.is_disabled
        assert related_rule2.rule.rule_number == 138
        assert related_rule2.rule.from_zone == "port3"
        assert related_rule2.rule.to_zone == "port2"
        assert related_rule2.rule.track["track_level"] == "LOG"
        assert related_rule2.rule.install_ons == []
        assert related_rule2.rule.communities == []
        assert related_rule2.rule.times == []

        assert len(related_rule2.rule.sources) == 1
        assert (
            related_rule2.rule.sources[0].uid
            == "{e08566b5-184e-463d-b6f5-de44aeede838}"
        )
        assert related_rule2.rule.sources[0].name == "Host_10.103.0.42"
        assert related_rule2.rule.sources[0].display_name == "Host_10.103.0.42"
        assert related_rule2.rule.sources[0].class_name.value == "host_plain"
        assert related_rule2.rule.sources[0].origin == "DEVICE"
        assert not related_rule2.rule.sources[0].is_global
        assert not related_rule2.rule.sources[0].is_implicit
        assert not related_rule2.rule.sources[0].is_shared
        assert related_rule2.rule.sources[0].comment == ""
        assert related_rule2.rule.sources[0].id == "66c7421a9591a744946348e4"
        assert related_rule2.rule.sources[0].version_id == 53
        assert related_rule2.rule.sources[0].device_type.value == "Fortinet"
        assert related_rule2.rule.sources[0].ip_type.value == "IPV4"
        assert related_rule2.rule.sources[0].referenced == "UNKNOWN"
        assert not related_rule2.rule.sources[0].installable_target

        assert len(related_rule2.rule.destinations) == 1
        assert (
            related_rule2.rule.destinations[0].uid
            == "{2c8c67e6-d1c2-48ee-a821-3e79b736d2d7}"
        )
        assert related_rule2.rule.destinations[0].name == "Host_10.234.10.42"
        assert related_rule2.rule.destinations[0].display_name == "Host_10.234.10.42"
        assert related_rule2.rule.destinations[0].class_name.value == "host_plain"
        assert related_rule2.rule.destinations[0].origin == "DEVICE"
        assert not related_rule2.rule.destinations[0].is_global
        assert not related_rule2.rule.destinations[0].is_implicit
        assert not related_rule2.rule.destinations[0].is_shared
        assert related_rule2.rule.destinations[0].comment == ""
        assert related_rule2.rule.destinations[0].id == "66c7421a9591a744946348e0"
        assert related_rule2.rule.destinations[0].version_id == 53
        assert related_rule2.rule.destinations[0].device_type.value == "Fortinet"
        assert related_rule2.rule.destinations[0].ip_type.value == "IPV4"
        assert related_rule2.rule.destinations[0].referenced == "UNKNOWN"
        assert not related_rule2.rule.destinations[0].installable_target

        assert len(related_rule2.rule.destination_services) == 1
        assert (
            related_rule2.rule.destination_services[0].uid
            == "{e28c04b8-cb4b-4f8f-a733-e95e5f654f0f}"
        )
        assert related_rule2.rule.destination_services[0].name == "TCP_442"
        assert related_rule2.rule.destination_services[0].display_name == "TCP_442"
        assert related_rule2.rule.destination_services[0].origin == "DEVICE"
        assert not related_rule2.rule.destination_services[0].is_global
        assert not related_rule2.rule.destination_services[0].implicit
        assert not related_rule2.rule.destination_services[0].shared
        assert related_rule2.rule.destination_services[0].timeout == 0
        assert related_rule2.rule.destination_services[0].comment == ""
        assert related_rule2.rule.destination_services[0].version_id == 53
        assert related_rule2.rule.destination_services[0].referenced.value == "UNKNOWN"
        assert related_rule2.rule.destination_services[0].type_on_device == ""
        assert not related_rule2.rule.destination_services[0].negate
        assert not related_rule2.rule.destination_services[0].match_for_any

        assert (
            related_rule2.intersecting_objects.found_in_source.uid
            == "{e08566b5-184e-463d-b6f5-de44aeede838}"
        )
        assert (
            related_rule2.intersecting_objects.found_in_destination.uid
            == "{2c8c67e6-d1c2-48ee-a821-3e79b736d2d7}"
        )
        assert (
            related_rule2.intersecting_objects.found_in_service.uid
            == "{e28c04b8-cb4b-4f8f-a733-e95e5f654f0f}"
        )



from enum import Enum
from typing import Optional

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, stringify_optional_obj

from netaddr import IPAddress


@propify
class GenericVpn(Jsonable):
    id: Optional[int] = prop(None, converter=optional(int))
    device_id: Optional[int] = prop(None, converter=optional(int), key="deviceId")
    generic: Optional[bool] = prop(None)
    vpn_name: Optional[str] = prop(None, key="vpnName")
    interface_name: Optional[str] = prop(None, key="interfaceName")
    tunnel_source_ip_addr: Optional[IPAddress] = prop(
        None,
        repr=False,
        jsonify=stringify_optional_obj,
        converter=optional(IPAddress),
        key="tunnelSourceIpAddr",
    )
    tunnel_dest_ip_addr: Optional[IPAddress] = prop(
        None,
        repr=False,
        jsonify=stringify_optional_obj,
        converter=optional(IPAddress),
        key="tunnelDestIpAddr",
    )



from copy import deepcopy
from datetime import datetime
from enum import Enum
from typing import List, Optional, Union, Type, Iterator
from typing_extensions import Literal
from time import sleep

# from requests import Response
# from requests.exceptions import HTTPError
# import attr

# from traversify import Traverser

# avoid circular imports
import pytos2  # noqa
from pytos2.utils import (
    propify,
    kwargify,
    prop,
    jsonify,
    safe_date,
    safe_iso8601_date,
    get_api_node,
    TimeFormat,
    safe_unwrap_msg,
)

from pytos2.models import Jsonable, UnMapped


@propify
class Query(Jsonable):

    class Meta(Enum):
        ROOT = "query"

    class Prop(Enum):
        LAST_USED = "lastUsed"
        CREATE_DATE = "createDate"
        UPDATE_DATE = "updateDate"

    type: str = prop(None)
    name: str = prop(None)
    description: Optional[str] = prop(None)
    last_used: Optional[datetime] = prop(
        None,
        key=Prop.LAST_USED.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )
    create_date: Optional[datetime] = prop(
        None,
        key=Prop.CREATE_DATE.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )
    update_date: Optional[datetime] = prop(
        None,
        key=Prop.UPDATE_DATE.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )


@propify
class ExpirationFilter(Jsonable):

    class Meta(Enum):
        ROOT = "expiration"

    class Prop(Enum):
        FROM_DATE = "fromDate"
        TO_DATE = "toDate"

    from_date: Optional[datetime] = prop(
        None,
        key=Prop.FROM_DATE.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )
    to_date: Optional[datetime] = prop(
        None,
        key=Prop.TO_DATE.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )


@propify
class FieldFilter(Jsonable):

    class Meta(Enum):
        ROOT = "field"

    name: Optional[str] = prop(None)
    value: Optional[str] = prop(None)


@propify
class DetailedQueryFilters(Jsonable):

    class Meta(Enum):
        ROOT = "filters"

    class Priority(Enum):
        CRITICAL = "Critical"
        HIGH = "High"
        NORMAL = "Normal"
        LOW = "Low"

    class TicketStatus(Enum):
        IN_PROGRESS = "IN_PROGRESS"
        REJECTED = "REJECTED"
        CLOSED = "CLOSED"
        CANCELLED = "CANCELLED"
        RESOLVED = "RESOLVED"

    class SlaStatus(Enum):
        OK = "OK"
        WARNING = "WARNING"
        ALERT = "ALERT"
        ESCALATION = "ESCALATION"
        NA = "NA"

    class SlaOutcome(Enum):
        UNKNOWN = "UNKNOWN"
        MET = "MET"
        OVERDUE = "OVERDUE"
        NA = "NA"

    class TaskStatus(Enum):
        WAITING_TO_BE_ASSIGNED = "WAITING_TO_BE_ASSIGNED"
        ASSIGNED = "ASSIGNED"
        WAITING_FOR_MORE_INFO = "WAITING_FOR_MORE_INFO"
        DONE = "DONE"
        INVALID = "INVALID"
        PENDING = "PENDING"
        PENDING_LICENSE = "PENDING_LICENSE"

    class Prop(Enum):
        DOMAIN_ID = "domainIds"
        TASK_STATUS = "taskStatuses"
        TICKET_STATUS = "ticketStatuses"
        SLA_STATUS = "slaStatuses"
        SLA_OUTCOME = "slaOutcomes"
        ASSIGNED_TO = "assignedTo"
        CURRENT_STEP_NAME = "currentStepName"

    #    def __post_init__(self):
    #        from pytos2.securechange.ticket import Ticket  # Local import to avoid circular dependency
    #        self.priorities: List[Ticket.Priority] = prop(factory=list)
    #        self.slaStatuses: List[Ticket.SlaStatus] = prop(factory=list, kwargify=str)

    #    from pytos2.securechange.ticket import Ticket
    #    priorities: List[Ticket.Priority] = prop(factory=list, kwargify=str)

    expiration: Optional[ExpirationFilter] = prop(None)
    domain_id: List[int] = prop(factory=list, key=Prop.DOMAIN_ID.value)
    group: Optional[str] = prop(None)
    ticket_id: Optional[int] = prop(None)
    priorities: List[Priority] = prop(factory=list, kwargify=str)
    subject: Optional[str] = prop(None)
    current_step_name: Optional[str] = prop(None, key=Prop.CURRENT_STEP_NAME.value)
    ticket_statuses: List[TicketStatus] = prop(
        factory=list, key=Prop.TICKET_STATUS.value, kwargify=str
    )
    task_status: List[TaskStatus] = prop(
        factory=list, key=Prop.TASK_STATUS.value, kwargify=str
    )
    sla_status: List[SlaStatus] = prop(
        factory=list, key=Prop.SLA_STATUS.value, kwargify=str
    )
    sla_outcome: List[SlaOutcome] = prop(
        factory=list, key=Prop.SLA_OUTCOME.value, kwargify=str
    )
    requester: Optional[str] = prop(None)
    assigned_to: Optional[str] = prop(None, key=Prop.ASSIGNED_TO.value)
    field: Optional[FieldFilter] = prop(None)


@propify
class FreeTextQuery(Query):

    class Meta(Enum):
        ROOT = "freeTextQuery"

    class QueryType(Enum):
        FREE_TEXT = "FREE_TEXT"

    class Prop(Enum):
        LAST_USED = "lastUsed"
        CREATE_DATE = "createDate"
        UPDATE_DATE = "updateDate"
        SEARCH_TEXT = "searchText"

    description: Optional[str] = prop(None)
    create_date: Optional[datetime] = prop(
        None,
        key=Prop.CREATE_DATE.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )
    update_date: Optional[datetime] = prop(
        None,
        key=Prop.UPDATE_DATE.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )
    search_text: str = prop(None, key=Prop.SEARCH_TEXT.value)
    last_used: Optional[datetime] = prop(
        None,
        key=Prop.LAST_USED.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )
    name: str = prop(None)
    id: Optional[int] = prop(None)
    type: Optional[QueryType] = prop(None, kwargify=str)


@propify
class DetailedQuery(Query):

    class Meta(Enum):
        ROOT = "detailedQuery"

    class QueryType(Enum):
        DETAILED = "DETAILED"

    class Prop(Enum):
        LAST_USED = "lastUsed"
        CREATE_DATE = "createDate"
        UPDATE_DATE = "updateDate"

    description: Optional[str] = prop(None)
    filters: DetailedQueryFilters = prop(None)
    create_date: Optional[datetime] = prop(
        None,
        key=Prop.CREATE_DATE.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )
    update_date: Optional[datetime] = prop(
        None,
        key=Prop.UPDATE_DATE.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )
    last_used: Optional[datetime] = prop(
        None,
        key=Prop.LAST_USED.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )
    name: str = prop(None)
    id: Optional[int] = prop(None)
    type: Optional[QueryType] = prop(None, kwargify=str)
